# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
import os
from unittest import mock

from oslo_utils import uuidutils

from tacker import context
from tacker.sol_refactored.common import vnfd_utils
from tacker.sol_refactored.conductor import vnflcm_driver_v2
from tacker.sol_refactored.nfvo import nfvo_client
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects.v2 import fields
from tacker.tests import base


SAMPLE_VNFD_ID = "b1bb0ce7-ebca-4fa7-95ed-4840d7000000"
SAMPLE_FLAVOUR_ID = "simple"


# InstantiateVnfRequest example for instantiate grant test
_ext_vl_1 = {
    "id": uuidutils.generate_uuid(),
    "resourceId": 'net0_id',
    "extCps": [
        {
            "cpdId": "VDU1_CP1",
            "cpConfig": {
                "VDU1_CP1_1": {
                    "cpProtocolData": [{
                        "layerProtocol": "IP_OVER_ETHERNET",
                        "ipOverEthernet": {
                            "ipAddresses": [{
                                "type": "IPV4",
                                "numDynamicAddresses": 1}]}}]}
            }
        },
        {
            "cpdId": "VDU2_CP1",
            "cpConfig": {
                "VDU2_CP1_1": {
                    "cpProtocolData": [{
                        "layerProtocol": "IP_OVER_ETHERNET",
                        "ipOverEthernet": {
                            "ipAddresses": [{
                                "type": "IPV4",
                                "fixedAddresses": ["10.10.0.101"]}]}}]}
            }
        }
    ],
}
_ext_vl_2 = {
    "id": uuidutils.generate_uuid(),
    "resourceId": 'net1_id',
    "extCps": [
        {
            "cpdId": "VDU1_CP2",
            "cpConfig": {
                "VDU1_CP2_1": {
                    "cpProtocolData": [{
                        "layerProtocol": "IP_OVER_ETHERNET",
                        "ipOverEthernet": {
                            "ipAddresses": [{
                                "type": "IPV4",
                                "numDynamicAddresses": 1,
                                "subnetId": 'subnet1_id'}]}}]}
            }
        },
        {
            "cpdId": "VDU2_CP2",
            "cpConfig": {
                "VDU2_CP2_1": {
                    "cpProtocolData": [{
                        "layerProtocol": "IP_OVER_ETHERNET",
                        "ipOverEthernet": {
                            "ipAddresses": [{
                                "type": "IPV4",
                                "fixedAddresses": ["10.10.1.101"],
                                "subnetId": 'subnet1_id'}]}}]}
            }
        }
    ]
}
_inst_req_example = {
    "flavourId": "simple",
    "instantiationLevelId": "instantiation_level_2",
    "extVirtualLinks": [
        _ext_vl_1,
        _ext_vl_2
    ],
    "extManagedVirtualLinks": [
        {
            "id": uuidutils.generate_uuid(),
            "vnfVirtualLinkDescId": "internalVL1",
            "resourceId": 'net_mgmt_id'
        },
    ],
    "vimConnectionInfo": {
        "vim1": {
            "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
            "vimId": uuidutils.generate_uuid(),
            "interfaceInfo": {"endpoint": "http://localhost/identity/v3"},
            "accessInfo": {
                "username": "nfv_user",
                "region": "RegionOne",
                "password": "devstack",
                "project": "nfv",
                "projectDomain": "Default",
                "userDomain": "Default"
            }
        }
    }
}

#  ChangeExtVnfConnectivityRequest example for change_ext_conn grant test
_ext_vl_3 = {
    "id": uuidutils.generate_uuid(),
    "resourceId": 'net2_id',
    "extCps": [
        {
            "cpdId": "VDU1_CP2",
            "cpConfig": {
                "VDU1_CP2_1": {
                    "cpProtocolData": [{
                        "layerProtocol": "IP_OVER_ETHERNET",
                        "ipOverEthernet": {
                            "ipAddresses": [{
                                "type": "IPV4",
                                "numDynamicAddresses": 1,
                                "subnetId": 'subnet2_id'}]}}]}
            }
        },
        {
            "cpdId": "VDU2_CP1",
            "cpConfig": {
                "VDU2_CP1_1": {
                    "linkPortId": "link_port_id"
                }
            }
        }
    ],
    "extLinkPorts": [
        {
            "id": "link_port_id",
            "resourceHandle": {
                "resourceId": "res_id_VDU2_CP1"
            }
        }
    ]
}
_change_ext_conn_req_example = {
    "extVirtualLinks": [_ext_vl_3]
}

# instantiatedVnfInfo example for terminate/scale grant test
# NOTE:
# - some identifiers are modified to make check easy.
# - some attributes which are not related to make terminate/scale grant
#   request are omitted.
_inst_info_example = {
    "flavourId": "simple",
    "vnfState": "STARTED",
    "extCpInfo": [
        {
            "id": "90561570-264c-4472-b84f-1fff98513475",
            "cpdId": "VDU2_CP1",
            "cpConfigId": "VDU2_CP1_1",
            # "cpProtocolInfo": omitted
            "extLinkPortId": "ac27c99b-73c8-4e91-b730-90deade72af4",
            "associatedVnfcCpId": "be955786-a0c7-4b61-8cd8-9bb8bcb1c6e3"
        },
        {
            "id": "f9f4b4b2-50e2-4c73-b89b-e0665e65ffbe",
            "cpdId": "VDU2_CP2",
            "cpConfigId": "VDU2_CP2_1",
            # "cpProtocolInfo": omitted
            "extLinkPortId": "12567a13-9fbd-4803-ad9f-d94ced266cd8",
            "associatedVnfcCpId": "c54fa2fc-185a-49a7-bb89-f30f7c3be6a4"
        },
        {
            "id": "05474d0b-a1f7-4be5-b57e-ef6873e1f3b6",
            "cpdId": "VDU1_CP2",
            "cpConfigId": "VDU1_CP2_1",
            # "cpProtocolInfo": omitted
            "extLinkPortId": "aa6646da-2e59-4de9-9b72-c62e7c4d9142",
            "associatedVnfcCpId": "fdbb289f-87c8-40d0-bf06-da07b41ba124"
        },
        {
            "id": "42ede9a6-c2b8-4c0d-a337-26342ffb236c",
            "cpdId": "VDU1_CP1",
            "cpConfigId": "VDU1_CP1_1",
            # "cpProtocolInfo": omitted
            "extLinkPortId": "efd0eb4e-4e55-4ac8-8b9b-403ec79faf2d",
            "associatedVnfcCpId": "235f920c-8b49-4894-9c36-73f5a3b9f74d"
        },
        {
            "id": "4f0ab1ad-b7de-482f-a69b-1093c71d2ceb",
            "cpdId": "VDU1_CP2",
            "cpConfigId": "VDU1_CP2_1",
            # "cpProtocolInfo": omitted
            "extLinkPortId": "a6c4c043-e082-4873-a871-02467af66224",
            "associatedVnfcCpId": "e23d970d-9ea9-4c26-9d67-8f244383ea3c"
        },
        {
            "id": "7a7fa30f-a303-4856-bc8b-b836cb682892",
            "cpdId": "VDU1_CP1",
            "cpConfigId": "VDU1_CP1_1",
            # "cpProtocolInfo": omitted
            "extLinkPortId": "f58df4d9-08ff-41b7-ab73-95ebfb8103c4",
            "associatedVnfcCpId": "259c5895-7be6-4bed-8a94-221c41b3d08f"
        }
    ],
    "extVirtualLinkInfo": [
        {
            "id": "137bdf0b-835c-43f0-b0d2-5c002599118a",
            "resourceHandle": {
                "resourceId": "6f97f400-2861-482a-ba78-65b652aaf8fc"
            },
            "extLinkPorts": [
                {
                    "id": "ac27c99b-73c8-4e91-b730-90deade72af4",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU2_CP1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "90561570-264c-4472-b84f-1fff98513475"
                },
                {
                    "id": "efd0eb4e-4e55-4ac8-8b9b-403ec79faf2d",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU1_1_CP1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "42ede9a6-c2b8-4c0d-a337-26342ffb236c"
                },
                {
                    "id": "f58df4d9-08ff-41b7-ab73-95ebfb8103c4",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU1_2_CP1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "7a7fa30f-a303-4856-bc8b-b836cb682892"
                }
            ],
            # "currentVnfExtCpData": omitted
        },
        {
            "id": "d8141a5a-6b6e-4dab-9bf5-158f23a617d7",
            "resourceHandle": {
                "resourceId": "02bc95e0-3d43-4d11-83b8-f7b15d8661a9"
            },
            "extLinkPorts": [
                {
                    "id": "12567a13-9fbd-4803-ad9f-d94ced266cd8",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU2_CP2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "f9f4b4b2-50e2-4c73-b89b-e0665e65ffbe"
                },
                {
                    "id": "aa6646da-2e59-4de9-9b72-c62e7c4d9142",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU1_1_CP2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "05474d0b-a1f7-4be5-b57e-ef6873e1f3b6"
                },
                {
                    "id": "a6c4c043-e082-4873-a871-02467af66224",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU1_2_CP2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "4f0ab1ad-b7de-482f-a69b-1093c71d2ceb"
                }
            ],
            # "currentVnfExtCpData": omitted
        }
    ],
    "extManagedVirtualLinkInfo": [
        {
            "id": "bad53df7-f1fa-482d-91b1-caec382aeec2",
            "vnfVirtualLinkDescId": "internalVL1",
            "networkResource": {
                "resourceId": "56730009-169c-4f96-8141-828acf1ee067"
            },
            "vnfLinkPorts": [
                {
                    "id": "74f387fe-6355-4af3-adc7-cdb507d5fa5f",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU2_CP3",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "5b0b336b-c207-4fa8-8b41-a5ad87d85cd0",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "4064ec55-b862-4527-a911-8752d3aa765a",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU1_1_CP3",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "b0732fb7-a42a-4077-aebc-d22b67b64f13",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "f3c9e62d-0f31-4a36-bd99-eecd8def0871",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU1_2_CP3",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "9c65f67e-feb2-447c-b0e7-a4f896185b4f",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        }
    ],
    "vnfcResourceInfo": [
        {
            "id": "vnfc_res_info_id_VDU2",
            "vduId": "VDU2",
            "computeResource": {
                # "vimConnectionId": omitted
                "resourceId": "res_id_VDU2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "vnfcCpInfo": [
                {
                    "id": "be955786-a0c7-4b61-8cd8-9bb8bcb1c6e3",
                    "cpdId": "VDU2_CP1",
                    "vnfExtCpId": "90561570-264c-4472-b84f-1fff98513475"
                },
                {
                    "id": "c54fa2fc-185a-49a7-bb89-f30f7c3be6a4",
                    "cpdId": "VDU2_CP2",
                    "vnfExtCpId": "f9f4b4b2-50e2-4c73-b89b-e0665e65ffbe"
                },
                {
                    "id": "5b0b336b-c207-4fa8-8b41-a5ad87d85cd0",
                    "cpdId": "VDU2_CP3",
                    "vnfLinkPortId": "74f387fe-6355-4af3-adc7-cdb507d5fa5f"
                },
                {
                    "id": "2cb2b3a8-a7a0-41da-b3b8-4b82f576b090",
                    "cpdId": "VDU2_CP4",
                    "vnfLinkPortId": "8e01813f-35fc-4a35-8f64-0da08a45ea21"
                },
                {
                    "id": "39a7d895-3b19-4330-b6ec-ae3557ea9c01",
                    "cpdId": "VDU2_CP5",
                    "vnfLinkPortId": "4dd7cadd-b9a1-484f-b2f2-1ff50ef0d90f"
                }
            ]
        },
        {
            "id": "vnfc_res_info_id_VDU1_1",
            "vduId": "VDU1",
            "computeResource": {
                # "vimConnectionId": omitted
                "resourceId": "res_id_VDU1_1",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": ["2135b13c-e630-4700-8f8d-85b6e48f7871"],
            "vnfcCpInfo": [
                {
                    "id": "235f920c-8b49-4894-9c36-73f5a3b9f74d",
                    "cpdId": "VDU1_CP1",
                    "vnfExtCpId": "42ede9a6-c2b8-4c0d-a337-26342ffb236c"
                },
                {
                    "id": "fdbb289f-87c8-40d0-bf06-da07b41ba124",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "05474d0b-a1f7-4be5-b57e-ef6873e1f3b6"
                },
                {
                    "id": "b0732fb7-a42a-4077-aebc-d22b67b64f13",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "4064ec55-b862-4527-a911-8752d3aa765a"
                },
                {
                    "id": "a49f8fb8-6fd9-4e9f-a6dd-0d268e51c83c",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "1666a0f7-6a34-474e-87a2-07fb0c30ecdb"
                },
                {
                    "id": "33194d65-ecd6-48d9-8ef7-c15ce9fef46c",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "ace663cd-431b-402a-b2ae-d0824c996edb"
                }
            ]
        },
        {
            "id": "vnfc_res_info_id_VDU1_2",
            "vduId": "VDU1",
            "computeResource": {
                # "vimConnectionId": omitted
                "resourceId": "res_id_VDU1_2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": ["739f7012-7973-485b-b34f-b006bc336150"],
            "vnfcCpInfo": [
                {
                    "id": "259c5895-7be6-4bed-8a94-221c41b3d08f",
                    "cpdId": "VDU1_CP1",
                    # when extLinkPorts of extVirtualLinks specified, there is
                    # no vnfExtCpId nor vnfLinkPortId.
                },
                {
                    "id": "e23d970d-9ea9-4c26-9d67-8f244383ea3c",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "4f0ab1ad-b7de-482f-a69b-1093c71d2ceb"
                },
                {
                    "id": "9c65f67e-feb2-447c-b0e7-a4f896185b4f",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "f3c9e62d-0f31-4a36-bd99-eecd8def0871"
                },
                {
                    "id": "0716ac20-612a-4ac2-8c87-d83be31dd4b5",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "bce3159b-caca-45b7-8bb7-88015e951e56"
                },
                {
                    "id": "c9112298-61eb-4bba-b285-ed3419593b1b",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "e0f98917-70ff-4f79-8747-9d7fc22827a4"
                }
            ]
        }
    ],
    "vnfVirtualLinkResourceInfo": [
        {
            "id": "18bd0111-d5e1-4aa3-b2d8-5b89833c6351",
            "vnfVirtualLinkDescId": "internalVL3",
            "networkResource": {
                # "vimConnectionId": omitted
                "resourceId": "res_id_internalVL3",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPorts": [
                {
                    "id": "4dd7cadd-b9a1-484f-b2f2-1ff50ef0d90f",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU2_CP5",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "39a7d895-3b19-4330-b6ec-ae3557ea9c01",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "ace663cd-431b-402a-b2ae-d0824c996edb",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU1_1_CP5",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "33194d65-ecd6-48d9-8ef7-c15ce9fef46c",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "e0f98917-70ff-4f79-8747-9d7fc22827a4",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU1_2_CP5",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "c9112298-61eb-4bba-b285-ed3419593b1b",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        },
        {
            "id": "047aa313-b591-4529-aa98-cb8ce2b82e28",
            "vnfVirtualLinkDescId": "internalVL2",
            "networkResource": {
                # "vimConnectionId": omitted
                "resourceId": "res_id_internalVL2",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPorts": [
                {
                    "id": "8e01813f-35fc-4a35-8f64-0da08a45ea21",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU2_CP4",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "2cb2b3a8-a7a0-41da-b3b8-4b82f576b090",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "1666a0f7-6a34-474e-87a2-07fb0c30ecdb",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU1_1_CP4",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "a49f8fb8-6fd9-4e9f-a6dd-0d268e51c83c",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "bce3159b-caca-45b7-8bb7-88015e951e56",
                    "resourceHandle": {
                        # "vimConnectionId": omitted
                        "resourceId": "res_id_VDU1_2_CP4",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "0716ac20-612a-4ac2-8c87-d83be31dd4b5",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        }
    ],
    "virtualStorageResourceInfo": [
        {
            "id": "2135b13c-e630-4700-8f8d-85b6e48f7871",
            "virtualStorageDescId": "VirtualStorage",
            "storageResource": {
                # "vimConnectionId": omitted
                "resourceId": "res_id_VirtualStorage_1",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        },
        {
            "id": "739f7012-7973-485b-b34f-b006bc336150",
            "virtualStorageDescId": "VirtualStorage",
            "storageResource": {
                # "vimConnectionId": omitted
                "resourceId": "res_id_VirtualStorage_2",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        }
    ],
    "vnfcInfo": [
        {
            "id": "VDU2-vnfc_res_info_id_VDU2",
            "vduId": "VDU2",
            "vnfcResourceInfoId": "vnfc_res_info_id_VDU2",
            "vnfcState": "STARTED"
        },
        {
            "id": "VDU1-vnfc_res_info_id_VDU1_1",
            "vduId": "VDU1",
            "vnfcResourceInfoId": "vnfc_res_info_id_VDU1_1",
            "vnfcState": "STARTED"
        },
        {
            "id": "VDU1-vnfc_res_info_id_VDU1_2",
            "vduId": "VDU1",
            "vnfcResourceInfoId": "vnfc_res_info_id_VDU1_2",
            "vnfcState": "STARTED"
        }
    ]
}

# modify_info_process example
_modify_inst_example = {
    "vnfInstanceName": "instance_name",
    "vnfInstanceDescription": "description",
    "vnfdId": SAMPLE_VNFD_ID,
    "vnfProvider": "provider",
    "vnfProductName": "product name",
    "vnfSoftwareVersion": "software version",
    "vnfdVersion": "vnfd version",
    "vnfConfigurableProperties": {
        "vnfproperties": "example"
    },
    "metadata": {
        "metadata": "example",
    },
    "vimConnectionInfo": {
        "vim1": {
            "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
            "vimId": "464bc3b0-5af3-40ef-bfb6-e84eee444313",
            "interfaceInfo": {"endpoint": "http://localhost/identity/v3"},
            "accessInfo": {
                "username": "nfv_user",
                "region": "RegionOne",
                "password": "devstack",
                "project": "nfv",
                "projectDomain": "Default",
                "userDomain": "Default"
            },
            "extra": {
                "key": "value"
            }
        }
    }
}


_modify_vnfc_info_example = {
    "vnfcInfo": [
        {
            "id": "VDU1-vnfc_res_info_id_VDU1",
            "vnfcConfigurableProperties": {"key": "value"}
        },
        {
            "id": "VDU2-vnfc_res_info_id_VDU2",
            "vnfcConfigurableProperties": {"key": "value"}
        }
    ]
}


class TestVnfLcmDriverV2(base.BaseTestCase):

    def setUp(self):
        super(TestVnfLcmDriverV2, self).setUp()
        objects.register_all()
        self.driver = vnflcm_driver_v2.VnfLcmDriverV2()
        self.context = context.get_admin_context()

        cur_dir = os.path.dirname(__file__)
        sample_dir = os.path.join(cur_dir, "..", "samples")

        self.vnfd_1 = vnfd_utils.Vnfd(SAMPLE_VNFD_ID)
        self.vnfd_1.init_from_csar_dir(os.path.join(sample_dir, "sample1"))

    def _grant_req_links(self, lcmocc_id, inst_id):
        return {
            'vnfLcmOpOcc': {
                'href': '{}/vnflcm/v2/vnf_lcm_op_occs/{}'.format(
                    self.driver.endpoint, lcmocc_id)
            },
            'vnfInstance': {
                'href': '{}/vnflcm/v2/vnf_instances/{}'.format(
                    self.driver.endpoint, inst_id)
            }
        }

    @mock.patch.object(nfvo_client.NfvoClient, 'grant')
    def test_instantiate_grant(self, mocked_grant):
        # prepare
        req = objects.InstantiateVnfRequest.from_dict(_inst_req_example)
        inst = objects.VnfInstanceV2(
            # required fields
            id=uuidutils.generate_uuid(),
            vnfdId=SAMPLE_VNFD_ID,
            vnfProvider='provider',
            vnfProductName='product name',
            vnfSoftwareVersion='software version',
            vnfdVersion='vnfd version',
            instantiationState='NOT_INSTANTIATED'
        )
        lcmocc = objects.VnfLcmOpOccV2(
            # required fields
            id=uuidutils.generate_uuid(),
            operationState=fields.LcmOperationStateType.STARTING,
            stateEnteredTime=datetime.utcnow(),
            startTime=datetime.utcnow(),
            vnfInstanceId=inst.id,
            operation=fields.LcmOperationType.INSTANTIATE,
            isAutomaticInvocation=False,
            isCancelPending=False,
            operationParams=req)

        mocked_grant.return_value = objects.GrantV1()

        # run instantiate_grant
        grant_req, _ = self.driver.grant(
            self.context, lcmocc, inst, self.vnfd_1)

        # check grant_req is constructed according to intention
        grant_req = grant_req.to_dict()
        expected_fixed_items = {
            'vnfInstanceId': inst.id,
            'vnfLcmOpOccId': lcmocc.id,
            'vnfdId': SAMPLE_VNFD_ID,
            'flavourId': SAMPLE_FLAVOUR_ID,
            'operation': 'INSTANTIATE',
            'isAutomaticInvocation': False,
            'instantiationLevelId': 'instantiation_level_2',
            '_links': self._grant_req_links(lcmocc.id, inst.id)
        }
        for key, value in expected_fixed_items.items():
            self.assertEqual(value, grant_req[key])

        add_reses = grant_req['addResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []},
            'STORAGE': {'VirtualStorage': []},
            'LINKPORT': {'VDU1_CP1': [], 'VDU1_CP2': [], 'VDU1_CP3': [],
                         'VDU1_CP4': [], 'VDU1_CP5': [],
                         'VDU2_CP1': [], 'VDU2_CP2': [], 'VDU2_CP3': [],
                         'VDU2_CP4': [], 'VDU2_CP5': []},
            # internalVL1 does not exist
            'VL': {'internalVL2': [], 'internalVL3': []}
        }
        expected_num = {
            'COMPUTE': {'VDU1': 3, 'VDU2': 1},
            'STORAGE': {'VirtualStorage': 3},
            'LINKPORT': {'VDU1_CP1': 3, 'VDU1_CP2': 3, 'VDU1_CP3': 3,
                         'VDU1_CP4': 3, 'VDU1_CP5': 3,
                         'VDU2_CP1': 1, 'VDU2_CP2': 1, 'VDU2_CP3': 1,
                         'VDU2_CP4': 1, 'VDU2_CP5': 1},
            'VL': {'internalVL2': 1, 'internalVL3': 1}
        }
        for res in add_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['id'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_num[key][name], len(ids))

        expected_placement_constraints = [{
            'affinityOrAntiAffinity': 'ANTI_AFFINITY',
            'scope': 'NFVI_NODE',
            'resource': []}]
        vdu_def_ids = (check_reses['COMPUTE']['VDU1'] +
                       check_reses['COMPUTE']['VDU2'])
        for def_id in vdu_def_ids:
            expected_placement_constraints[0]['resource'].append(
                {'idType': 'GRANT', 'resourceId': def_id})
        self.assertEqual(expected_placement_constraints,
                         grant_req['placementConstraints'])

    @mock.patch.object(nfvo_client.NfvoClient, 'grant')
    def test_terminate_grant(self, mocked_grant):
        # prepare
        inst = objects.VnfInstanceV2(
            # required fields
            id=uuidutils.generate_uuid(),
            vnfdId=SAMPLE_VNFD_ID,
            vnfProvider='provider',
            vnfProductName='product name',
            vnfSoftwareVersion='software version',
            vnfdVersion='vnfd version',
            instantiationState='INSTANTIATED'
        )
        inst_info = objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
            _inst_info_example)
        inst.instantiatedVnfInfo = inst_info
        req = objects.TerminateVnfRequest.from_dict(
            {"terminationType": "FORCEFUL"})
        lcmocc = objects.VnfLcmOpOccV2(
            # required fields
            id=uuidutils.generate_uuid(),
            operationState=fields.LcmOperationStateType.STARTING,
            stateEnteredTime=datetime.utcnow(),
            startTime=datetime.utcnow(),
            vnfInstanceId=inst.id,
            operation=fields.LcmOperationType.TERMINATE,
            isAutomaticInvocation=False,
            isCancelPending=False,
            operationParams=req)

        mocked_grant.return_value = objects.GrantV1()

        # run terminate_grant
        grant_req, _ = self.driver.grant(
            self.context, lcmocc, inst, self.vnfd_1)

        # check grant_req is constructed according to intention
        grant_req = grant_req.to_dict()
        expected_fixed_items = {
            'vnfInstanceId': inst.id,
            'vnfLcmOpOccId': lcmocc.id,
            'vnfdId': SAMPLE_VNFD_ID,
            'operation': 'TERMINATE',
            'isAutomaticInvocation': False,
            '_links': self._grant_req_links(lcmocc.id, inst.id)
        }
        for key, value in expected_fixed_items.items():
            self.assertEqual(value, grant_req[key])

        rm_reses = grant_req['removeResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []},
            'STORAGE': {'VirtualStorage': []},
            'LINKPORT': {'VDU1_CP1': [], 'VDU1_CP2': [], 'VDU1_CP3': [],
                         'VDU1_CP4': [], 'VDU1_CP5': [],
                         'VDU2_CP1': [], 'VDU2_CP2': [], 'VDU2_CP3': [],
                         'VDU2_CP4': [], 'VDU2_CP5': []},
            'VL': {'internalVL2': [], 'internalVL3': []}
        }
        expected_res_ids = {
            'COMPUTE': {
                'VDU1': ['res_id_VDU1_1', 'res_id_VDU1_2'],
                'VDU2': ['res_id_VDU2']
            },
            'STORAGE': {
                'VirtualStorage': ['res_id_VirtualStorage_1',
                                   'res_id_VirtualStorage_2']
            },
            'LINKPORT': {
                'VDU1_CP1': ['res_id_VDU1_1_CP1'],
                'VDU1_CP2': ['res_id_VDU1_1_CP2', 'res_id_VDU1_2_CP2'],
                'VDU1_CP3': ['res_id_VDU1_1_CP3', 'res_id_VDU1_2_CP3'],
                'VDU1_CP4': ['res_id_VDU1_1_CP4', 'res_id_VDU1_2_CP4'],
                'VDU1_CP5': ['res_id_VDU1_1_CP5', 'res_id_VDU1_2_CP5'],
                'VDU2_CP1': ['res_id_VDU2_CP1'],
                'VDU2_CP2': ['res_id_VDU2_CP2'],
                'VDU2_CP3': ['res_id_VDU2_CP3'],
                'VDU2_CP4': ['res_id_VDU2_CP4'],
                'VDU2_CP5': ['res_id_VDU2_CP5']
            },
            'VL': {
                'internalVL2': ['res_id_internalVL2'],
                'internalVL3': ['res_id_internalVL3']
            }
        }
        for res in rm_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['resource']['resourceId'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_res_ids[key][name], ids)

    def _scale_grant_prepare(self, scale_type):
        inst = objects.VnfInstanceV2(
            # required fields
            id=uuidutils.generate_uuid(),
            vnfdId=SAMPLE_VNFD_ID,
            vnfProvider='provider',
            vnfProductName='product name',
            vnfSoftwareVersion='software version',
            vnfdVersion='vnfd version',
            instantiationState='INSTANTIATED'
        )
        inst_info = objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
            _inst_info_example)
        inst.instantiatedVnfInfo = inst_info
        req = objects.ScaleVnfRequest.from_dict(
            {"type": scale_type,
             "aspectId": "VDU1_scale",
             "numberOfSteps": 1})
        lcmocc = objects.VnfLcmOpOccV2(
            # required fields
            id=uuidutils.generate_uuid(),
            operationState=fields.LcmOperationStateType.STARTING,
            stateEnteredTime=datetime.utcnow(),
            startTime=datetime.utcnow(),
            vnfInstanceId=inst.id,
            operation=fields.LcmOperationType.SCALE,
            isAutomaticInvocation=False,
            isCancelPending=False,
            operationParams=req)

        return inst, lcmocc

    @mock.patch.object(nfvo_client.NfvoClient, 'grant')
    def test_scale_grant_scale_out(self, mocked_grant):
        # prepare
        inst, lcmocc = self._scale_grant_prepare('SCALE_OUT')
        mocked_grant.return_value = objects.GrantV1()

        # run scale_grant scale-out
        grant_req, _ = self.driver.grant(
            self.context, lcmocc, inst, self.vnfd_1)

        # check grant_req is constructed according to intention
        grant_req = grant_req.to_dict()
        expected_fixed_items = {
            'vnfInstanceId': inst.id,
            'vnfLcmOpOccId': lcmocc.id,
            'vnfdId': SAMPLE_VNFD_ID,
            'operation': 'SCALE',
            'isAutomaticInvocation': False,
            '_links': self._grant_req_links(lcmocc.id, inst.id)
        }
        for key, value in expected_fixed_items.items():
            self.assertEqual(value, grant_req[key])

        add_reses = grant_req['addResources']
        check_reses = {
            'COMPUTE': {'VDU1': []},
            'STORAGE': {'VirtualStorage': []},
            'LINKPORT': {'VDU1_CP1': [], 'VDU1_CP2': [], 'VDU1_CP3': [],
                         'VDU1_CP4': [], 'VDU1_CP5': []}
        }
        expected_num = {
            'COMPUTE': {'VDU1': 1},
            'STORAGE': {'VirtualStorage': 1},
            'LINKPORT': {'VDU1_CP1': 1, 'VDU1_CP2': 1, 'VDU1_CP3': 1,
                         'VDU1_CP4': 1, 'VDU1_CP5': 1}
        }
        for res in add_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['id'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_num[key][name], len(ids))

    @mock.patch.object(nfvo_client.NfvoClient, 'grant')
    def test_scale_grant_scale_in(self, mocked_grant):
        # prepare
        inst, lcmocc = self._scale_grant_prepare('SCALE_IN')
        mocked_grant.return_value = objects.GrantV1()

        # run scale_grant scale-in
        grant_req, _ = self.driver.grant(
            self.context, lcmocc, inst, self.vnfd_1)

        # check grant_req is constructed according to intention
        grant_req = grant_req.to_dict()
        expected_fixed_items = {
            'vnfInstanceId': inst.id,
            'vnfLcmOpOccId': lcmocc.id,
            'vnfdId': SAMPLE_VNFD_ID,
            'operation': 'SCALE',
            'isAutomaticInvocation': False,
            '_links': self._grant_req_links(lcmocc.id, inst.id)
        }
        for key, value in expected_fixed_items.items():
            self.assertEqual(value, grant_req[key])

        rm_reses = grant_req['removeResources']
        check_reses = {
            'COMPUTE': {'VDU1': []},
            'STORAGE': {'VirtualStorage': []},
            'LINKPORT': {'VDU1_CP1': [], 'VDU1_CP2': [], 'VDU1_CP3': [],
                         'VDU1_CP4': [], 'VDU1_CP5': []}
        }
        expected_res_ids = {
            'COMPUTE': {
                'VDU1': ['res_id_VDU1_1']
            },
            'STORAGE': {
                'VirtualStorage': ['res_id_VirtualStorage_1']
            },
            'LINKPORT': {
                'VDU1_CP1': ['res_id_VDU1_1_CP1'],
                'VDU1_CP2': ['res_id_VDU1_1_CP2'],
                'VDU1_CP3': ['res_id_VDU1_1_CP3'],
                'VDU1_CP4': ['res_id_VDU1_1_CP4'],
                'VDU1_CP5': ['res_id_VDU1_1_CP5']
            }
        }
        for res in rm_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['resource']['resourceId'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_res_ids[key][name], ids)

    def test_make_inst_info_common_instantiate(self):
        # prepare
        inst_saved = objects.VnfInstanceV2(
            # only set used members in the method
            instantiatedVnfInfo=objects.VnfInstanceV2_InstantiatedVnfInfo()
        )
        inst = inst_saved.obj_clone()
        req = objects.InstantiateVnfRequest.from_dict(_inst_req_example)
        lcmocc = objects.VnfLcmOpOccV2(
            # only set used members in the method
            operation=fields.LcmOperationType.INSTANTIATE,
            operationParams=req)

        # run _make_inst_info_common
        self.driver._make_inst_info_common(
            lcmocc, inst_saved, inst, self.vnfd_1)

        inst = inst.to_dict()
        expected_scale_status = [{'aspectId': 'VDU1_scale', 'scaleLevel': 2}]
        expected_max_scale_levels = [
            {'aspectId': 'VDU1_scale', 'scaleLevel': 2}]

        self.assertEqual(expected_scale_status,
                         inst['instantiatedVnfInfo']['scaleStatus'])
        self.assertEqual(expected_max_scale_levels,
                         inst['instantiatedVnfInfo']['maxScaleLevels'])

    def test_make_inst_info_common_scale(self):
        # prepare
        inst_saved = objects.VnfInstanceV2(
            # only set used members in the method
            instantiatedVnfInfo=objects.VnfInstanceV2_InstantiatedVnfInfo()
        )
        inst_saved.instantiatedVnfInfo.scaleStatus = [
            objects.ScaleInfoV2(aspectId='VDU1_scale', scaleLevel=2)
        ]
        inst_saved.instantiatedVnfInfo.maxScaleLevels = [
            objects.ScaleInfoV2(aspectId='VDU1_scale', scaleLevel=2)
        ]
        inst = objects.VnfInstanceV2(
            # only set used members in the method
            instantiatedVnfInfo=objects.VnfInstanceV2_InstantiatedVnfInfo()
        )
        req = objects.ScaleVnfRequest.from_dict(
            {"type": "SCALE_IN",
             "aspectId": "VDU1_scale",
             "numberOfSteps": 1})
        lcmocc = objects.VnfLcmOpOccV2(
            # only set used members in the method
            operation=fields.LcmOperationType.SCALE,
            operationParams=req)

        # run _make_inst_info_common
        self.driver._make_inst_info_common(
            lcmocc, inst_saved, inst, self.vnfd_1)

        inst = inst.to_dict()
        expected_scale_status = [{'aspectId': 'VDU1_scale', 'scaleLevel': 1}]
        expected_max_scale_levels = [
            {'aspectId': 'VDU1_scale', 'scaleLevel': 2}]

        self.assertEqual(expected_scale_status,
                         inst['instantiatedVnfInfo']['scaleStatus'])
        self.assertEqual(expected_max_scale_levels,
                         inst['instantiatedVnfInfo']['maxScaleLevels'])

    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnf_package_info_vnfd')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnfd_utils.Vnfd, 'get_vnfd_properties')
    def test_modify_info_process(self, mocked_get_vnfd_properties,
            mocked_get_vnfd, mocked_get_vnf_package_info_vnfd):
        new_vnfd_id = uuidutils.generate_uuid()
        pkg_info = objects.VnfPkgInfoV2(
            id=uuidutils.generate_uuid(),
            vnfdId=new_vnfd_id,
            vnfProvider="provider_1",
            vnfProductName="product_1",
            vnfSoftwareVersion="software version",
            vnfdVersion="vnfd version",
            operationalState="ENABLED"
        )
        new_vnfd_prop = {
            "vnfConfigurableProperties": {
                "vnfproperties": "example"
            },
            "metadata": {
                "metadata": "example",
                "metadata_2": "example"
            },
            "extensions": {
                "extensions": "example"
            },
        }
        mocked_get_vnf_package_info_vnfd.return_value = pkg_info
        mocked_get_vnfd.return_value = vnfd_utils.Vnfd(new_vnfd_id)
        mocked_get_vnfd_properties.return_value = new_vnfd_prop
        req = objects.VnfInfoModificationRequest.from_dict(
            {
                "vnfInstanceName": "instance_name",
                "vnfInstanceDescription": "description_1",
                "vnfdId": new_vnfd_id,
                "vnfConfigurableProperties": {
                    "vnfproperties": "example"
                },
                "metadata": {
                    "metadata_1": "example_1",
                    "metadata_2": None
                },
                "extensions": {
                    "extensions": "example_1"
                },
                "vimConnectionInfo": {
                    "vim1": {
                        "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
                        "vimId": "464bc3b0-5af3-40ef-bfb6-e84eee444313",
                        "interfaceInfo": {
                            "endpoint": "http://localhost/identity/v3"
                        },
                        "accessInfo": {
                            "username": "nfv_user",
                            "region": "RegionOne",
                            "password": "devstack",
                            "project": "nfv",
                            "projectDomain": "Default",
                            "userDomain": "Default"
                        },
                        "extra": {
                            "key_add": "value"
                        }
                    }
                },
                "vnfcInfoModifications": [
                    {
                        "id": "VDU1-vnfc_res_info_id_VDU1",
                        "vnfcConfigurableProperties": {
                            "key": "value_mod",
                            "key_add": "value"
                        }
                    },
                    {
                        "id": "VDU2-vnfc_res_info_id_VDU2",
                        "vnfcConfigurableProperties": {
                            "key": None
                        }
                    }
                ]
            }
        )
        inst = objects.VnfInstanceV2.from_dict(_modify_inst_example)
        vnfc_info = objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
            _modify_vnfc_info_example)
        inst.instantiatedVnfInfo = vnfc_info

        lcmocc = objects.VnfLcmOpOccV2(
            # required fields
            id=uuidutils.generate_uuid(),
            operationState=fields.LcmOperationStateType.PROCESSING,
            stateEnteredTime=datetime.utcnow(),
            startTime=datetime.utcnow(),
            operation=fields.LcmOperationType.MODIFY_INFO,
            isAutomaticInvocation=False,
            isCancelPending=False,
            operationParams=req)

        # run modify_info_process
        self.driver.modify_info_process(
            self.context, lcmocc, inst, None, None, self.vnfd_1)

        inst = inst.to_dict()
        expected_modify_result = {
            "vnfInstanceName": "instance_name",
            "vnfInstanceDescription": "description_1",
            "vnfdId": new_vnfd_id,
            "vnfProvider": "provider_1",
            "vnfProductName": "product_1",
            "vnfSoftwareVersion": "software version",
            "vnfdVersion": "vnfd version",
            "vnfConfigurableProperties": {
                "vnfproperties": "example"
            },
            "metadata": {
                "metadata": "example",
                "metadata_1": "example_1"
            },
            "extensions": {
                "extensions": "example_1"
            },
            "vimConnectionInfo": {
                "vim1": {
                    "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
                    "vimId": "464bc3b0-5af3-40ef-bfb6-e84eee444313",
                    "interfaceInfo": {
                        "endpoint": "http://localhost/identity/v3"
                    },
                    "accessInfo": {
                        "username": "nfv_user",
                        "region": "RegionOne",
                        "password": "devstack",
                        "project": "nfv",
                        "projectDomain": "Default",
                        "userDomain": "Default"
                    },
                    "extra": {
                        "key": "value",
                        "key_add": "value"
                    }
                }
            },
            "instantiatedVnfInfo": {
                "vnfcInfo": [
                    {
                        "id": "VDU1-vnfc_res_info_id_VDU1",
                        "vnfcConfigurableProperties": {
                            "key": "value_mod",
                            "key_add": "value"
                        }
                    },
                    {
                        "id": "VDU2-vnfc_res_info_id_VDU2",
                        "vnfcConfigurableProperties": {
                        }
                    }
                ]
            }
        }

        self.assertEqual(expected_modify_result, inst)

    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnf_package_info_vnfd')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnfd_utils.Vnfd, 'get_vnfd_properties')
    def test_modify_info_process_from_vnfd_prop(self,
            mocked_get_vnfd_properties, mocked_get_vnfd,
            mocked_get_vnf_package_info_vnfd):
        new_vnfd_id = uuidutils.generate_uuid()
        pkg_info = objects.VnfPkgInfoV2(
            id=uuidutils.generate_uuid(),
            vnfdId=new_vnfd_id,
            vnfProvider="provider_1",
            vnfProductName="product_1",
            vnfSoftwareVersion="software version",
            vnfdVersion="vnfd version",
            operationalState="ENABLED"
        )
        new_vnfd_prop = {}

        mocked_get_vnf_package_info_vnfd.return_value = pkg_info
        mocked_get_vnfd.return_value = vnfd_utils.Vnfd(new_vnfd_id)
        mocked_get_vnfd_properties.return_value = new_vnfd_prop
        req = objects.VnfInfoModificationRequest.from_dict(
            {
                "vnfInstanceName": "instance_name",
                "vnfInstanceDescription": "description_1",
                "vnfdId": new_vnfd_id,
                "vimConnectionInfo": {
                    "vim1": {
                        "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
                        "vimId": "464bc3b0-5af3-40ef-bfb6-e84eee444313",
                        "interfaceInfo": {
                            "endpoint": "http://localhost/identity/v3"
                        },
                        "accessInfo": {
                            "username": "nfv_user",
                            "region": "RegionOne",
                            "password": "devstack",
                            "project": "nfv",
                            "projectDomain": "Default",
                            "userDomain": "Default"
                        }
                    }
                }
            }
        )
        inst = objects.VnfInstanceV2.from_dict(_modify_inst_example)
        lcmocc = objects.VnfLcmOpOccV2(
            # required fields
            id=uuidutils.generate_uuid(),
            operationState=fields.LcmOperationStateType.PROCESSING,
            stateEnteredTime=datetime.utcnow(),
            startTime=datetime.utcnow(),
            operation=fields.LcmOperationType.MODIFY_INFO,
            isAutomaticInvocation=False,
            isCancelPending=False,
            operationParams=req)

        # run modify_info_process
        self.driver.modify_info_process(
            self.context, lcmocc, inst, None, None, self.vnfd_1)

        inst = inst.to_dict()
        expected_modify_result = {
            "vnfInstanceName": "instance_name",
            "vnfInstanceDescription": "description_1",
            "vnfdId": new_vnfd_id,
            "vnfProvider": "provider_1",
            "vnfProductName": "product_1",
            "vnfSoftwareVersion": "software version",
            "vnfdVersion": "vnfd version",
            "vnfConfigurableProperties": {},
            "metadata": {},
            "vimConnectionInfo": {
                "vim1": {
                    "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
                    "vimId": "464bc3b0-5af3-40ef-bfb6-e84eee444313",
                    "interfaceInfo": {
                        "endpoint": "http://localhost/identity/v3"
                    },
                    "accessInfo": {
                        "username": "nfv_user",
                        "region": "RegionOne",
                        "password": "devstack",
                        "project": "nfv",
                        "projectDomain": "Default",
                        "userDomain": "Default"
                    },
                    "extra": {
                        "key": "value"
                    }
                }
            }
        }

        self.assertEqual(expected_modify_result, inst)

    @mock.patch.object(nfvo_client.NfvoClient, 'grant')
    def test_change_ext_conn_grant(self, mocked_grant):
        # prepare
        req = objects.ChangeExtVnfConnectivityRequest.from_dict(
            _change_ext_conn_req_example)
        inst = objects.VnfInstanceV2(
            # required fields
            id=uuidutils.generate_uuid(),
            vnfdId=SAMPLE_VNFD_ID,
            vnfProvider='provider',
            vnfProductName='product name',
            vnfSoftwareVersion='software version',
            vnfdVersion='vnfd version',
            instantiationState='INSTANTIATED'
        )
        inst_info = objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
            _inst_info_example)
        inst.instantiatedVnfInfo = inst_info
        lcmocc = objects.VnfLcmOpOccV2(
            # required fields
            id=uuidutils.generate_uuid(),
            operationState=fields.LcmOperationStateType.STARTING,
            stateEnteredTime=datetime.utcnow(),
            startTime=datetime.utcnow(),
            vnfInstanceId=inst.id,
            operation=fields.LcmOperationType.CHANGE_EXT_CONN,
            isAutomaticInvocation=False,
            isCancelPending=False,
            operationParams=req
        )

        mocked_grant.return_value = objects.GrantV1()

        # run change_ext_conn_grant
        grant_req, _ = self.driver.grant(
            self.context, lcmocc, inst, self.vnfd_1)

        # check grant_req is constructed according to intention
        grant_req = grant_req.to_dict()
        expected_fixed_items = {
            'vnfInstanceId': inst.id,
            'vnfLcmOpOccId': lcmocc.id,
            'vnfdId': SAMPLE_VNFD_ID,
            'operation': 'CHANGE_EXT_CONN',
            'isAutomaticInvocation': False,
            '_links': self._grant_req_links(lcmocc.id, inst.id)
        }
        for key, value in expected_fixed_items.items():
            self.assertEqual(value, grant_req[key])

        # check updateResources
        update_reses = grant_req['updateResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []}
        }
        expected_res_ids = {
            'COMPUTE': {
                'VDU1': ['res_id_VDU1_1', 'res_id_VDU1_2'],
                'VDU2': ['res_id_VDU2']
            }
        }
        for res in update_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['resource']['resourceId'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_res_ids[key][name], ids)

        # check removeResources
        rm_reses = grant_req['removeResources']
        check_reses = {
            'LINKPORT': {'VDU1_CP2': [], 'VDU2_CP1': []}
        }
        expected_res_ids = {
            'LINKPORT': {
                'VDU1_CP2': ['res_id_VDU1_1_CP2', 'res_id_VDU1_2_CP2'],
                'VDU2_CP1': ['res_id_VDU2_CP1']
            }
        }
        for res in rm_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['resource']['resourceId'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_res_ids[key][name], ids)

        # check addResources
        add_reses = grant_req['addResources']
        check_reses = {
            'LINKPORT': {'VDU1_CP2': []}
        }
        expected_num = {
            'LINKPORT': {'VDU1_CP2': 2}
        }
        for res in add_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['id'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_num[key][name], len(ids))

    def _heal_grant_prepare(self, req):
        inst = objects.VnfInstanceV2(
            # required fields
            id=uuidutils.generate_uuid(),
            vnfdId=SAMPLE_VNFD_ID,
            vnfProvider='provider',
            vnfProductName='product name',
            vnfSoftwareVersion='software version',
            vnfdVersion='vnfd version',
            instantiationState='INSTANTIATED'
        )
        inst_info = objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
            _inst_info_example)
        inst.instantiatedVnfInfo = inst_info
        lcmocc = objects.VnfLcmOpOccV2(
            # required fields
            id=uuidutils.generate_uuid(),
            operationState=fields.LcmOperationStateType.STARTING,
            stateEnteredTime=datetime.utcnow(),
            startTime=datetime.utcnow(),
            vnfInstanceId=inst.id,
            operation=fields.LcmOperationType.HEAL,
            isAutomaticInvocation=False,
            isCancelPending=False,
            operationParams=req)

        return inst, lcmocc

    @mock.patch.object(nfvo_client.NfvoClient, 'grant')
    def test_heal_grant_SOL002(self, mocked_grant):
        # prepare
        req = objects.HealVnfRequest(
            vnfcInstanceId=["VDU1-vnfc_res_info_id_VDU1_2",
                            "VDU2-vnfc_res_info_id_VDU2"]
        )
        inst, lcmocc = self._heal_grant_prepare(req)
        mocked_grant.return_value = objects.GrantV1()

        # run heal_grant
        grant_req, _ = self.driver.grant(
            self.context, lcmocc, inst, self.vnfd_1)

        # check grant_req is constructed according to intention
        grant_req = grant_req.to_dict()
        expected_fixed_items = {
            'vnfInstanceId': inst.id,
            'vnfLcmOpOccId': lcmocc.id,
            'vnfdId': SAMPLE_VNFD_ID,
            'operation': 'HEAL',
            'isAutomaticInvocation': False,
            '_links': self._grant_req_links(lcmocc.id, inst.id)
        }
        for key, value in expected_fixed_items.items():
            self.assertEqual(value, grant_req[key])

        # check removeResources
        rm_reses = grant_req['removeResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []}
        }
        expected_res_ids = {
            'COMPUTE': {
                'VDU1': ['res_id_VDU1_2'],
                'VDU2': ['res_id_VDU2']
            }
        }
        for res in rm_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['resource']['resourceId'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_res_ids[key][name], ids)

        # check addResources
        add_reses = grant_req['addResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []}
        }
        for res in add_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['id'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(len(expected_res_ids[key][name]), len(ids))

    @mock.patch.object(nfvo_client.NfvoClient, 'grant')
    def test_heal_grant_SOL002_all(self, mocked_grant):
        # prepare
        req = objects.HealVnfRequest(
            vnfcInstanceId=["VDU1-vnfc_res_info_id_VDU1_2",
                            "VDU2-vnfc_res_info_id_VDU2"],
            additionalParams={'all': True}
        )
        inst, lcmocc = self._heal_grant_prepare(req)
        mocked_grant.return_value = objects.GrantV1()

        # run heal_grant
        grant_req, _ = self.driver.grant(
            self.context, lcmocc, inst, self.vnfd_1)

        # check grant_req is constructed according to intention
        grant_req = grant_req.to_dict()
        expected_fixed_items = {
            'vnfInstanceId': inst.id,
            'vnfLcmOpOccId': lcmocc.id,
            'vnfdId': SAMPLE_VNFD_ID,
            'operation': 'HEAL',
            'isAutomaticInvocation': False,
            '_links': self._grant_req_links(lcmocc.id, inst.id)
        }
        for key, value in expected_fixed_items.items():
            self.assertEqual(value, grant_req[key])

        # check removeResources
        rm_reses = grant_req['removeResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []},
            'STORAGE': {'VirtualStorage': []}
        }
        expected_res_ids = {
            'COMPUTE': {
                'VDU1': ['res_id_VDU1_2'],
                'VDU2': ['res_id_VDU2']
            },
            'STORAGE': {
                'VirtualStorage': ['res_id_VirtualStorage_2']
            }
        }
        for res in rm_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['resource']['resourceId'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_res_ids[key][name], ids)

        # check addResources
        add_reses = grant_req['addResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []},
            'STORAGE': {'VirtualStorage': []}
        }
        for res in add_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['id'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(len(expected_res_ids[key][name]), len(ids))

    @mock.patch.object(nfvo_client.NfvoClient, 'grant')
    def test_heal_grant_SOL003(self, mocked_grant):
        # prepare
        req = objects.HealVnfRequest()
        inst, lcmocc = self._heal_grant_prepare(req)
        mocked_grant.return_value = objects.GrantV1()

        # run heal_grant
        grant_req, _ = self.driver.grant(
            self.context, lcmocc, inst, self.vnfd_1)

        # check grant_req is constructed according to intention
        grant_req = grant_req.to_dict()
        expected_fixed_items = {
            'vnfInstanceId': inst.id,
            'vnfLcmOpOccId': lcmocc.id,
            'vnfdId': SAMPLE_VNFD_ID,
            'operation': 'HEAL',
            'isAutomaticInvocation': False,
            '_links': self._grant_req_links(lcmocc.id, inst.id)
        }
        for key, value in expected_fixed_items.items():
            self.assertEqual(value, grant_req[key])

        # check removeResources
        rm_reses = grant_req['removeResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []}
        }
        expected_res_ids = {
            'COMPUTE': {
                'VDU1': ['res_id_VDU1_1', 'res_id_VDU1_2'],
                'VDU2': ['res_id_VDU2']
            }
        }
        for res in rm_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['resource']['resourceId'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_res_ids[key][name], ids)

        # check addResources
        add_reses = grant_req['addResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []}
        }
        for res in add_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['id'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(len(expected_res_ids[key][name]), len(ids))

    @mock.patch.object(nfvo_client.NfvoClient, 'grant')
    def test_heal_grant_SOL003_all(self, mocked_grant):
        # prepare
        req = objects.HealVnfRequest(
            additionalParams={'all': True}
        )
        inst, lcmocc = self._heal_grant_prepare(req)
        mocked_grant.return_value = objects.GrantV1()

        # run heal_grant
        grant_req, _ = self.driver.grant(
            self.context, lcmocc, inst, self.vnfd_1)

        # check grant_req is constructed according to intention
        grant_req = grant_req.to_dict()
        expected_fixed_items = {
            'vnfInstanceId': inst.id,
            'vnfLcmOpOccId': lcmocc.id,
            'vnfdId': SAMPLE_VNFD_ID,
            'operation': 'HEAL',
            'isAutomaticInvocation': False,
            '_links': self._grant_req_links(lcmocc.id, inst.id)
        }
        for key, value in expected_fixed_items.items():
            self.assertEqual(value, grant_req[key])

        # check removeResources
        rm_reses = grant_req['removeResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []},
            'STORAGE': {'VirtualStorage': []},
            'LINKPORT': {'VDU1_CP1': [], 'VDU1_CP2': [], 'VDU1_CP3': [],
                         'VDU1_CP4': [], 'VDU1_CP5': [],
                         'VDU2_CP1': [], 'VDU2_CP2': [], 'VDU2_CP3': [],
                         'VDU2_CP4': [], 'VDU2_CP5': []},
            'VL': {'internalVL2': [], 'internalVL3': []}
        }
        expected_res_ids = {
            'COMPUTE': {
                'VDU1': ['res_id_VDU1_1', 'res_id_VDU1_2'],
                'VDU2': ['res_id_VDU2']
            },
            'STORAGE': {
                'VirtualStorage': ['res_id_VirtualStorage_1',
                                   'res_id_VirtualStorage_2']
            },
            'LINKPORT': {
                'VDU1_CP1': ['res_id_VDU1_1_CP1'],
                'VDU1_CP2': ['res_id_VDU1_1_CP2', 'res_id_VDU1_2_CP2'],
                'VDU1_CP3': ['res_id_VDU1_1_CP3', 'res_id_VDU1_2_CP3'],
                'VDU1_CP4': ['res_id_VDU1_1_CP4', 'res_id_VDU1_2_CP4'],
                'VDU1_CP5': ['res_id_VDU1_1_CP5', 'res_id_VDU1_2_CP5'],
                'VDU2_CP1': ['res_id_VDU2_CP1'],
                'VDU2_CP2': ['res_id_VDU2_CP2'],
                'VDU2_CP3': ['res_id_VDU2_CP3'],
                'VDU2_CP4': ['res_id_VDU2_CP4'],
                'VDU2_CP5': ['res_id_VDU2_CP5']
            },
            'VL': {
                'internalVL2': ['res_id_internalVL2'],
                'internalVL3': ['res_id_internalVL3']
            }
        }
        for res in rm_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['resource']['resourceId'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(expected_res_ids[key][name], ids)

        # check addResources
        add_reses = grant_req['addResources']
        check_reses = {
            'COMPUTE': {'VDU1': [], 'VDU2': []},
            'STORAGE': {'VirtualStorage': []},
            'LINKPORT': {'VDU1_CP1': [], 'VDU1_CP2': [], 'VDU1_CP3': [],
                         'VDU1_CP4': [], 'VDU1_CP5': [],
                         'VDU2_CP1': [], 'VDU2_CP2': [], 'VDU2_CP3': [],
                         'VDU2_CP4': [], 'VDU2_CP5': []},
            'VL': {'internalVL2': [], 'internalVL3': []}
        }
        for res in add_reses:
            check_reses[res['type']][res['resourceTemplateId']].append(
                res['id'])

        for key, value in check_reses.items():
            for name, ids in value.items():
                self.assertEqual(len(expected_res_ids[key][name]), len(ids))



# Copyright (C) 2020 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""change type for vnf_resources and vnf_instantiated_info table

Revision ID: aaf461c8844c
Revises: 745e3e9fe5e2
Create Date: 2020-09-17 03:17:42.570250

"""
# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = 'aaf461c8844c'
down_revision = '745e3e9fe5e2'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.alter_column('vnf_instantiated_info',
                    'instance_id',
                    type_=sa.Text(),
                    nullable=True)
    op.alter_column('vnf_resources',
                    'resource_name',
                    type_=sa.Text(),
                    nullable=True)



#    Copyright 2019 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_serialization import jsonutils as json
from oslo_utils import excutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
from oslo_utils import versionutils
from oslo_versionedobjects import base as ovoo_base
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import func

from tacker._i18n import _
from tacker.common import exceptions
from tacker.common import utils
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker import objects
from tacker.objects import base
from tacker.objects import common
from tacker.objects import fields
from tacker.objects import vnf_artifact
from tacker.objects import vnf_software_image

_NO_DATA_SENTINEL = object()

VNF_PACKAGE_OPTIONAL_ATTRS = [
    'vnf_deployment_flavours',
    'vnfd',
    'vnf_artifacts']

LOG = logging.getLogger(__name__)


def _add_user_defined_data(context, package_uuid, user_data,
                           max_retries=10):
    for attempt in range(max_retries):
        try:
            with db_api.context_manager.writer.using(context):
                new_entries = []
                for key, value in user_data.items():
                    new_entries.append({"key": key,
                                        "value": value,
                                        "package_uuid": package_uuid})
                if new_entries:
                    context.session.execute(
                        models.VnfPackageUserData.__table__.insert(None),
                        new_entries)

                return user_data
        except db_exc.DBDuplicateEntry:
            # a concurrent transaction has been committed,
            # try again unless this was the last attempt
            with excutils.save_and_reraise_exception() as context:
                if attempt < max_retries - 1:
                    context.reraise = False
                else:
                    raise exceptions.UserDataUpdateCreateFailed(
                        id=package_uuid, retries=max_retries)


def _vnf_package_user_data_get_query(context, package_uuid, model):
    return api.model_query(context, model, read_deleted="no",
                           project_only=True).\
        filter_by(package_uuid=package_uuid)


@db_api.context_manager.writer
def _update_user_defined_data(context, package_uuid, user_data):
    model = models.VnfPackageUserData
    user_data = user_data.copy()
    session = context.session
    with session.begin(subtransactions=True):
        # Get existing user_data
        db_user_data = _vnf_package_user_data_get_query(context, package_uuid,
                                                        model).all()
        save = []
        skip = []
        # We only want to send changed user_data.
        for row in db_user_data:
            if row.key in user_data:
                value = user_data.pop(row.key)
                if row.value != value:
                    # ORM objects will not be saved until we do the bulk save
                    row.value = value
                    save.append(row)
                    continue
            skip.append(row)

        # We also want to save non-existent user_data
        save.extend(model(key=key, value=value, package_uuid=package_uuid)
                    for key, value in user_data.items())
        # Do a bulk save
        if save:
            session.bulk_save_objects(save, update_changed_only=True)

        # Construct result dictionary with current user_data
        save.extend(skip)
        result = {row['key']: row['value'] for row in save}

    return result


@db_api.context_manager.reader
def _vnf_package_get_by_id(context, package_uuid, columns_to_join=None):

    query = api.model_query(context, models.VnfPackage,
                            read_deleted="no", project_only=True). \
        filter_by(id=package_uuid).options(joinedload('_metadata'))

    if columns_to_join:
        for column in columns_to_join:
            query = query.options(joinedload(column))

    result = query.first()

    if not result:
        raise exceptions.VnfPackageNotFound(id=package_uuid)

    return result


@db_api.context_manager.writer
def _vnf_package_create(context, values, user_data=None):

    vnf_package = models.VnfPackage()
    vnf_package.update(values)
    vnf_package.save(context.session)
    vnf_package._metadata = []

    if user_data:
        _add_user_defined_data(context, vnf_package.id, user_data)
        context.session.expire(vnf_package, ['_metadata'])
        vnf_package._metadata

    return vnf_package


@db_api.context_manager.reader
def _vnf_package_list(context, columns_to_join=None):
    query = api.model_query(context, models.VnfPackage, read_deleted="no",
                            project_only=True).options(joinedload('_metadata'))

    if columns_to_join:
        for column in columns_to_join:
            query = query.options(joinedload(column))

    return query.all()


@db_api.context_manager.reader
def _vnf_package_list_by_filters(context, read_deleted=None, filters=None):
    query = api.model_query(context, models.VnfPackage,
                            read_deleted=read_deleted,
                            project_only=True).options(joinedload('_metadata'))

    if filters:
        # Need to join VnfDeploymentFlavour, VnfSoftwareImage and
        # VnfSoftwareImageMetadata db table explicitly
        # only when filters contains one of the column matching
        # from VnfSoftwareImage or VnfSoftwareImageMetadata db table.
        filter_data = json.dumps(filters)
        if 'VnfSoftwareImageMetadata' in filter_data:
            query = query.join(models.VnfDeploymentFlavour).join(
                models.VnfSoftwareImage).join(
                models.VnfSoftwareImageMetadata)
        elif 'VnfSoftwareImage' in filter_data:
            query = query.join(models.VnfDeploymentFlavour).join(
                models.VnfSoftwareImage)

        if 'VnfPackageArtifactInfo' in filter_data:
            query = query.join(models.VnfPackageArtifactInfo)

        if 'VnfPackageVnfd' in filter_data:
            query = query.join(models.VnfPackageVnfd)

        query = common.apply_filters(query, filters)

    return query.all()


@db_api.context_manager.writer
def _update_vnf_package_except_user_data(context, vnf_package):
    vnf_package.save(session=context.session)


def _vnf_package_update(context, package_uuid, values, columns_to_join=None):
    user_data = values.pop('user_data', None)
    if user_data:
        _update_user_defined_data(context, package_uuid, user_data)

    vnf_package = _vnf_package_get_by_id(
        context, package_uuid, columns_to_join=columns_to_join)
    if values:
        vnf_package.update(values)
        _update_vnf_package_except_user_data(context, vnf_package)

    return vnf_package


@db_api.context_manager.writer
def _destroy_vnf_package(context, package_uuid):
    now = timeutils.utcnow()
    updated_values = {'deleted': True,
                      'deleted_at': now
                      }

    flavour_query = api.model_query(
        context, models.VnfDeploymentFlavour,
        (models.VnfDeploymentFlavour.id, )).filter_by(
        package_uuid=package_uuid)

    software_images_query = api.model_query(
        context, models.VnfSoftwareImage,
        (models.VnfSoftwareImage.id, )).filter(
        models.VnfSoftwareImage.flavour_uuid.in_(flavour_query.subquery()))

    api.model_query(
        context, models.VnfSoftwareImageMetadata).filter(
        models.VnfSoftwareImageMetadata.image_uuid.in_(
            software_images_query.subquery())).update(
        updated_values, synchronize_session=False)

    api.model_query(context, models.VnfSoftwareImage). \
        filter(models.VnfSoftwareImage.id.in_(
            flavour_query.subquery())).update(
        updated_values, synchronize_session=False)

    api.model_query(context, models.VnfPackageUserData). \
        filter_by(package_uuid=package_uuid). \
        update(updated_values, synchronize_session=False)
    api.model_query(context, models.VnfDeploymentFlavour). \
        filter_by(package_uuid=package_uuid). \
        update(updated_values, synchronize_session=False)
    api.model_query(context, models.VnfPackageArtifactInfo). \
        filter_by(package_uuid=package_uuid). \
        update(updated_values, synchronize_session=False)
    api.model_query(context, models.VnfPackageVnfd). \
        filter_by(package_uuid=package_uuid). \
        soft_delete(synchronize_session=False)
    api.model_query(context, models.VnfPackage).\
        filter_by(id=package_uuid). \
        update(updated_values, synchronize_session=False)


def _make_vnf_packages_list(context, vnf_package_list, db_vnf_package_list,
                            expected_attrs=None):
    vnf_package_cls = VnfPackage

    vnf_package_list.objects = []
    for db_package in db_vnf_package_list:
        vnf_pkg_obj = vnf_package_cls._from_db_object(
            context, vnf_package_cls(context), db_package,
            expected_attrs=expected_attrs)
        vnf_package_list.objects.append(vnf_pkg_obj)

    vnf_package_list.obj_reset_changes()
    return vnf_package_list


@base.TackerObjectRegistry.register
class VnfPackage(base.TackerObject, base.TackerPersistentObject,
                 base.TackerObjectDictCompat):

    # Key corresponds to the name of the parameter as defined
    # in type VnfPkgInfo of SOL003 document and value will contain tuple
    # of following values:-
    # 1. Parameter that is mapped to version object
    # 2. Data type of the field as per the data types supported by
    # attribute-based filtering
    # 3. DB model that's mapped to the version object.
    # 4. Valid values for a given data type if any. This value is set
    # especially for 'enum' data type.
    ALL_ATTRIBUTES = {
        'id': ('id', "uuid", 'VnfPackage'),
        'onboardingState': ('onboarding_state', "enum", 'VnfPackage',
            fields.PackageOnboardingStateTypeField().valid_values),
        'operationalState': ('operational_state', 'enum', 'VnfPackage',
            fields.PackageOperationalStateTypeField().valid_values),
        'usageState': ('usage_state', 'enum', 'VnfPackage',
            fields.PackageUsageStateTypeField().valid_values),
        'vnfProvider': ('vnfd.vnf_provider', 'string', 'VnfPackageVnfd'),
        'vnfProductName': ('vnfd.vnf_product_name', 'string',
                           'VnfPackageVnfd'),
        'vnfdId': ('vnfd.vnfd_id', 'uuid', 'VnfPackageVnfd'),
        'vnfSoftwareVersion': ('vnfd.vnf_software_version', 'string',
                               'VnfPackageVnfd'),
        'vnfdVersion': ('vnfd.vnfd_version', 'string', 'VnfPackageVnfd'),
        'userDefinedData/*': ('user_data', 'key_value_pair',
            {"key_column": "key", "value_column": "value",
             "model": "VnfPackageUserData"}),
        "checksum": {
            'algorithm': ('algorithm', 'string', 'VnfPackage'),
            'hash': ('hash', 'string', 'VnfPackage'),
        }
    }

    ALL_ATTRIBUTES.update(vnf_software_image.VnfSoftwareImage.ALL_ATTRIBUTES)
    ALL_ATTRIBUTES.update(vnf_artifact.VnfPackageArtifactInfo.ALL_ATTRIBUTES)

    FLATTEN_ATTRIBUTES = utils.flatten_dict(ALL_ATTRIBUTES.copy())

    simple_attributes = ['id', 'onboardingState', 'operationalState',
                         'usageState']
    simple_instantiated_attributes = ['vnfProvider', 'vnfProductName',
                         'vnfdId', 'vnfSoftwareVersion', 'vnfdVersion']

    COMPLEX_ATTRIBUTES = ["checksum", "userDefinedData"]
    COMPLEX_ATTRIBUTES.extend(
        vnf_software_image.VnfSoftwareImage.COMPLEX_ATTRIBUTES)
    COMPLEX_ATTRIBUTES.extend(vnf_artifact.VnfPackageArtifactInfo.
                              COMPLEX_ATTRIBUTES)

    # Version 1.1: Added 'size' to persist size of VnfPackage.
    VERSION = '1.1'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'onboarding_state': fields.PackageOnboardingStateTypeField(
            nullable=False),
        'operational_state': fields.PackageOperationalStateTypeField(
            nullable=False),
        'usage_state': fields.PackageUsageStateTypeField(nullable=False),
        'user_data': fields.DictOfStringsField(),
        'tenant_id': fields.StringField(nullable=False),
        'algorithm': fields.StringField(nullable=True),
        'hash': fields.StringField(nullable=True),
        'location_glance_store': fields.StringField(nullable=True),
        'vnf_deployment_flavours': fields.ObjectField(
            'VnfDeploymentFlavoursList', nullable=True),
        'vnfd': fields.ObjectField('VnfPackageVnfd', nullable=True),
        'size': fields.IntegerField(nullable=False, default=0),
        'vnf_artifacts': fields.ObjectField('VnfPackageArtifactInfoList',
                                            nullable=True)
    }

    def __init__(self, context=None, **kwargs):
        super(VnfPackage, self).__init__(context, **kwargs)
        self.obj_set_defaults()

    def obj_make_compatible(self, primitive, target_version):
        super(VnfPackage, self).obj_make_compatible(primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1) and 'size' in primitive:
            del primitive['size']

    @staticmethod
    def _from_db_object(context, vnf_package, db_vnf_package,
                        expected_attrs=None):
        expected_attrs = expected_attrs or []

        vnf_package._context = context

        for key in vnf_package.fields:
            if key in VNF_PACKAGE_OPTIONAL_ATTRS:
                continue
            if key == 'user_data':
                db_key = 'metadetails'
            else:
                db_key = key
            setattr(vnf_package, key, db_vnf_package[db_key])

        vnf_package._context = context
        vnf_package._extra_attributes_from_db_object(
            vnf_package, db_vnf_package, expected_attrs)

        vnf_package.obj_reset_changes()
        return vnf_package

    @staticmethod
    def _extra_attributes_from_db_object(vnf_package, db_vnf_package,
                                         expected_attrs=None):
        """Method to help with migration of extra attributes to objects."""

        if expected_attrs is None:
            expected_attrs = []

        if 'vnf_deployment_flavours' in expected_attrs:
            vnf_package._load_vnf_deployment_flavours(
                db_vnf_package.get('vnf_deployment_flavours'))

        if 'vnfd' in expected_attrs:
            vnf_package._load_vnfd(db_vnf_package.get('vnfd'))

        if 'vnf_artifacts' in expected_attrs:
            vnf_package._load_vnf_artifacts(
                db_vnf_package.get('vnf_artifacts'))

    def _load_vnf_deployment_flavours(self, db_flavours=_NO_DATA_SENTINEL):
        if db_flavours is _NO_DATA_SENTINEL:
            vnf_package = self.get_by_id(
                self._context, self.id,
                expected_attrs=['vnf_deployment_flavours'])
            if 'vnf_deployment_flavours' in vnf_package:
                self.vnf_deployment_flavours = \
                    vnf_package.vnf_deployment_flavours
                self.vnf_deployment_flavours.obj_reset_changes(recursive=True)
                self.obj_reset_changes(['vnf_deployment_flavours'])
            else:
                self.vnf_deployment_flavours = \
                    objects.VnfDeploymentFlavoursList(objects=[])
        elif db_flavours:
            self.vnf_deployment_flavours = base.obj_make_list(
                self._context, objects.VnfDeploymentFlavoursList(
                    self._context), objects.VnfDeploymentFlavour, db_flavours)
            self.obj_reset_changes(['vnf_deployment_flavours'])

    def _load_vnfd(self, db_vnfd=_NO_DATA_SENTINEL):
        if db_vnfd is None:
            self.vnfd = None
        elif db_vnfd is _NO_DATA_SENTINEL:
            vnf_package = self.get_by_id(self._context, self.id,
                                         expected_attrs=['vnfd'])

            if 'vnfd' in vnf_package and vnf_package.vnfd is not None:
                self.vnfd = vnf_package.vnfd
                self.vnfd.obj_reset_changes(recursive=True)
                self.obj_reset_changes(['vnfd'])
            else:
                self.vnfd = None
        elif db_vnfd:
            self.vnfd = objects.VnfPackageVnfd.obj_from_db_obj(
                self._context, db_vnfd)
            self.obj_reset_changes(['vnfd'])

    def _load_vnf_artifacts(self, db_artifact=_NO_DATA_SENTINEL):
        if db_artifact is _NO_DATA_SENTINEL:
            vnf_package = self.get_by_id(
                self._context, self.id,
                expected_attrs=['vnf_artifacts'])
            if 'vnf_artifacts' in vnf_package:
                self.vnf_artifacts = vnf_package.vnf_artifacts
                self.vnf_artifacts.obj_reset_changes(recursive=True)
                self.obj_reset_changes(['vnf_artifacts'])
            else:
                self.vnf_artifacts = objects.\
                    VnfPackageArtifactInfoList(objects=[])
        elif db_artifact:
            self.vnf_artifacts = base.obj_make_list(
                self._context, objects.VnfPackageArtifactInfoList(
                    self._context), objects.VnfPackageArtifactInfo,
                db_artifact)
            self.obj_reset_changes(['vnf_artifacts'])

    def _load_generic(self, attrname):
        vnf_package = self.__class__.get_by_id(self._context,
                                               id=self.id,
                                               expected_attrs=None)
        if attrname not in vnf_package:
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('loading %s requires recursion') % attrname)

        for field in self.fields:
            if field in vnf_package and field not in self:
                setattr(self, field, getattr(vnf_package, field))

    def obj_load_attr(self, attrname):
        if not self._context:
            raise exceptions.OrphanedObjectError(
                method='obj_load_attr', objtype=self.obj_name())
        if 'id' not in self:
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)

        LOG.debug("Lazy-loading '%(attr)s' on %(name)s id %(id)s",
                  {'attr': attrname,
                   'name': self.obj_name(),
                   'id': self.id,
                   })

        self._obj_load_attr(attrname)

    def _obj_load_attr(self, attrname):
        """Internal method for loading attributes from vnf package."""

        if attrname == 'vnf_deployment_flavours':
            self._load_vnf_deployment_flavours()
        elif attrname == 'vnfd':
            self._load_vnfd()
        elif attrname == 'vnf_artifacts':
            self._load_vnf_artifacts()
        elif attrname in self.fields and attrname != 'id':
            self._load_generic(attrname)
        else:
            # NOTE(nirajsingh): Raise error if non existing field is
            # requested.
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)

        self.obj_reset_changes([attrname])

    @base.remotable
    def create(self):
        updates = self.obj_get_changes()

        if 'id' not in updates:
            updates['id'] = uuidutils.generate_uuid()
            self.id = updates['id']

        for key in ['vnf_deployment_flavours']:
            if key in updates.keys():
                updates.pop(key)

        for key in ['vnf_artifacts']:
            if key in updates.keys():
                updates.pop(key)

        user_data = updates.pop('user_data', None)
        db_vnf_package = _vnf_package_create(self._context, updates,
                                             user_data=user_data)
        self._from_db_object(self._context, self, db_vnf_package)

    @base.remotable_classmethod
    def get_by_id(cls, context, id, expected_attrs=None):
        db_vnf_package = _vnf_package_get_by_id(
            context, id, columns_to_join=expected_attrs)
        return cls._from_db_object(context, cls(), db_vnf_package,
                                   expected_attrs=expected_attrs)

    @base.remotable
    def destroy(self, context):
        if not self.obj_attr_is_set('id'):
            raise exceptions.ObjectActionError(action='destroy',
                                               reason='no uuid')

        _destroy_vnf_package(context, self.id)

    @base.remotable
    def save(self):
        updates = self.tacker_obj_get_changes()
        for key in ['vnf_deployment_flavours']:
            if key in updates.keys():
                updates.pop(key)

        db_vnf_package = _vnf_package_update(self._context,
                                            self.id, updates)
        self._from_db_object(self._context, self, db_vnf_package)

    @base.remotable
    def is_package_in_use(self, context):
        if self.onboarding_state == \
                fields.PackageOnboardingStateType.ONBOARDED:
            # check if vnf package is used by any vnf instances.
            query = context.session.query(
                func.count(models.VnfInstance.id)).\
                filter_by(tenant_id=self.tenant_id).\
                filter_by(vnfd_id=self.vnfd.vnfd_id).\
                filter_by(deleted=False)
            result = query.scalar()
            return True if result > 0 else False
        else:
            return False

    def _get_vnfd(self, include_fields=None):
        response = dict()
        to_fields = set(self.simple_instantiated_attributes).intersection(
            include_fields)
        for field in to_fields:
            response[field] = utils.deepgetattr(self,
                self.FLATTEN_ATTRIBUTES[field][0])
        return response

    def _get_checksum(self, include_fields=None):
        response = dict()
        to_fields = set([key for key in self.FLATTEN_ATTRIBUTES.keys()
            if key.startswith('checksum')])
        to_fields = to_fields.intersection(include_fields)
        for field in to_fields:
            display_field = field.split("/")[-1]
            response[display_field] = getattr(self,
                self.FLATTEN_ATTRIBUTES[field][0])
        return {'checksum': response} if response else None

    def _get_user_defined_data(self, include_fields=None):
        # Need special handling for field containing key-value pair.
        # If user requests userDefined/key1 and if userDefineData contains
        # key1=value1, key2-value2, it should return only keys that are
        # requested in include_fields. If user requests only userDefinedData,
        # then in that case,it should return all key/value pairs. In case,
        # if any of the requested key is not present, then it should
        # siliently ignore it.
        key = 'userDefinedData'
        if key in include_fields or 'userDefinedData/*' in include_fields:
            return {key: self.user_data}
        else:
            # Check if user has requested specified keys from
            # userDefinedData.
            data_resp = dict()
            key_list = []
            special_key = 'userDefinedData/'
            for field in include_fields:
                if field.startswith(special_key):
                    key_list.append(field[len(special_key):])

            for key_req in key_list:
                if key_req in self.user_data:
                    data_resp[key_req] = self.user_data[key_req]

            if data_resp:
                return {key: data_resp}

    def _basic_vnf_package_info(self, include_fields=None):
        response = dict()
        to_fields = set(self.simple_attributes).intersection(include_fields)
        for field in to_fields:
            response[field] = getattr(self, self.FLATTEN_ATTRIBUTES[field][0])
        return response

    def to_dict(self, include_fields=None):
        if not include_fields:
            include_fields = set(self.FLATTEN_ATTRIBUTES.keys())

        vnf_package_response = self._basic_vnf_package_info(
            include_fields=include_fields)

        user_defined_data = self._get_user_defined_data(
            include_fields=include_fields)

        if user_defined_data:
            vnf_package_response.update(user_defined_data)

        if (self.onboarding_state ==
                fields.PackageOnboardingStateType.ONBOARDED):

            software_images = self.vnf_deployment_flavours.to_dict(
                include_fields=include_fields)
            if software_images:
                vnf_package_response.update(
                    {'softwareImages': software_images})

            vnf_package_response.update(self._get_vnfd(
                include_fields=include_fields))

            checksum = self._get_checksum(include_fields=include_fields)
            if checksum:
                vnf_package_response.update(checksum)

            artifacts = self.vnf_artifacts.to_dict(
                include_fields=include_fields)
            if artifacts:
                vnf_package_response.update(
                    {'additionalArtifacts': artifacts})

        return vnf_package_response


@base.TackerObjectRegistry.register
class VnfPackagesList(ovoo_base.ObjectListBase, base.TackerObject):

    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('VnfPackage')
    }

    @base.remotable_classmethod
    def get_all(cls, context, expected_attrs=None):
        db_vnf_packages = _vnf_package_list(context,
                                            columns_to_join=expected_attrs)
        return _make_vnf_packages_list(context, cls(), db_vnf_packages,
                                       expected_attrs)

    @base.remotable_classmethod
    def get_by_filters(cls, context, read_deleted=None, filters=None):
        db_vnf_packages = _vnf_package_list_by_filters(context,
                                            read_deleted=read_deleted,
                                            filters=filters)
        return _make_vnf_packages_list(context, cls(), db_vnf_packages)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from wsgiref.simple_server import make_server


PORT = 6789


def handle_notification(environ, start_response):
    method = environ['REQUEST_METHOD']
    print("notification %s" % method)
    if method not in ['GET', 'POST']:
        print("  not support method")
        start_response('405 not suportted method',
                       [('Content-Type', 'application/problem+json')])
        problem_detail = {'status': 405,
                          'detail': "not supported method"}
        body = json.dumps(problem_detail)
        return [body.encode('utf-8')]

    authorization = environ.get("HTTP_AUTHORIZATION", "")
    version = environ.get("HTTP_VERSION", "")
    print("  authorizarion: %s" % authorization)
    print("  version: %s" % version)

    if method == 'POST':
        length = environ.get('CONTENT_LENGTH')
        print("  content_length: %s" % length)
        body = environ.get('wsgi.input').read(int(length))
        body = json.loads(body.decode('utf-8'))
        print("  request body: %s" % body)

    start_response('204 No content', [])
    return []


def handle_token(environ, start_response):
    method = environ['REQUEST_METHOD']
    print("token %s" % method)
    if method not in ['POST']:
        print("  not support method")
        start_response('405 not suportted method',
                       [('Content-Type', 'application/problem+json')])
        problem_detail = {'status': 405,
                          'detail': "not supported method"}
        body = json.dumps(problem_detail)
        return [body.encode('utf-8')]

    authorization = environ.get("HTTP_AUTHORIZATION", "")
    version = environ.get("HTTP_VERSION", "")
    content_type = environ.get("CONTENT_TYPE")
    print("  authorizarion: %s" % authorization)
    print("  version: %s" % version)
    print("  content_type: %s" % content_type)

    length = environ.get('CONTENT_LENGTH')
    print("  content_length: %s" % length)
    body = environ.get('wsgi.input').read(int(length))
    body = body.decode('utf-8')
    print("  request body: %s" % body)

    item = body.split('=')
    if (len(item) != 2 or
            item[0] != 'grant_type' or
            item[1] != 'client_credentials'):
        start_response('400 Bad Request', [])
        return []

    start_response('200 OK', [('Content-Type', 'application/json')])
    data = {
        "access_token": "2YotnFZFEjr1zCsicMWpAA",
        "token_type": "example",
        "expires_in": 3600,
        "example_parameter": "example_value"
    }
    body = json.dumps(data)
    return [body.encode('utf-8')]


def notif_endpoint_app(environ, start_response):
    path = environ['PATH_INFO']

    if path == "/notification":
        return handle_notification(environ, start_response)

    if path == "/token":
        return handle_token(environ, start_response)


if __name__ == '__main__':
    try:
        with make_server('', PORT, notif_endpoint_app) as httpd:
            httpd.serve_forever()
            httpd.handle_request()
    except KeyboardInterrupt:
        print()
        print("End.")



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""rpc_proxy

Revision ID: 81ffa86020d
Revises: 1c6b0d82afcd
Create Date: 2014-03-19 15:50:11.712686

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '81ffa86020d'
down_revision = '1c6b0d82afcd'

# Change to ['*'] if this migration applies to all plugins

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.create_table(
        'proxymgmtports',
        sa.Column('device_id', sa.String(255)),
        sa.Column('port_id', sa.String(36), nullable=False),
        sa.Column('dst_transport_url', sa.String(255)),
        sa.Column('svr_proxy_id', sa.String(36)),
        sa.Column('svr_ns_proxy_id', sa.String(36)),
        sa.Column('clt_proxy_id', sa.String(36)),
        sa.Column('clt_ns_proxy_id', sa.String(36)),
        sa.PrimaryKeyConstraint('device_id'),
    )
    op.create_table(
        'proxyserviceports',
        sa.Column('service_instance_id', sa.String(255)),
        sa.Column('svr_proxy_id', sa.String(36)),
        sa.Column('svr_ns_proxy_id', sa.String(36)),
        sa.Column('clt_proxy_id', sa.String(36)),
        sa.Column('clt_ns_proxy_id', sa.String(36)),
        sa.PrimaryKeyConstraint('service_instance_id'),
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from heatclient import exc as heatException
from oslo_log import log as logging
from oslo_utils import timeutils

from tacker.common import clients
from tacker.extensions import vnfm

LOG = logging.getLogger(__name__)


class HeatClient(object):
    def __init__(self, auth_attr, region_name=None):
        # context, password are unused
        self.heat = clients.OpenstackClients(auth_attr, region_name).heat
        self.stacks = self.heat.stacks
        self.resource_types = self.heat.resource_types
        self.resources = self.heat.resources

    def _stack_ids(self, stack_id):
        filters = {"owner_id": stack_id,
            "show_nested": True}

        for stack in self.stacks.list(**{"filters": filters}):
            yield stack.id
            if stack.parent and stack.parent == stack_id:
                for x in self._stack_ids(stack.id):
                    yield x

    def find_stack(self, **kwargs):
        stack = None
        for sub_stack in self.stacks.list(**{"filters": kwargs}):
            stack = sub_stack
        return stack

    def get_cooldown(self, stack_id, rsc_name):
        metadata = self.resource_metadata(stack_id, rsc_name)
        if 'cooldown_end' in metadata:
            now = timeutils.utcnow()
            cooldown_end = timeutils.parse_strtime(
                next(iter(metadata['cooldown_end'].keys())))
            return max(0, timeutils.delta_seconds(now, cooldown_end))
        return None

    def create(self, fields):
        fields = fields.copy()
        fields['disable_rollback'] = True
        if 'password' in fields.get('template', {}):
            fields['password'] = fields['template']['password']

        fields.setdefault('timeout_mins')

        try:
            return self.stacks.create(**fields)
        except heatException.HTTPException:
            type_, value, tb = sys.exc_info()
            raise vnfm.HeatClientException(msg=value)

    def delete(self, stack_id):
        try:
            self.stacks.delete(stack_id)
        except heatException.HTTPNotFound:
            LOG.warning("Stack %(stack)s created by service chain driver is "
                        "not found at cleanup", {'stack': stack_id})

    def get(self, stack_id):
        return self.stacks.get(stack_id)

    def get_stack_nested_depth(self, stack_id):
        stack_ids = self._stack_ids(stack_id)
        if stack_ids:
            return len(list(stack_ids))

        return 0

    def update(self, stack_id, **kwargs):
        try:
            return self.stacks.update(stack_id, **kwargs)
        except heatException.HTTPException:
            type_, value, tb = sys.exc_info()
            raise vnfm.HeatClientException(msg=value)

    def resource_attr_support(self, resource_name, property_name):
        resource = self.resource_types.get(resource_name)
        return property_name in resource['attributes']

    def resource_get_list(self, stack_id, nested_depth=0):
        return self.heat.resources.list(stack_id,
                                        nested_depth=nested_depth)

    def resource_signal(self, stack_id, rsc_name):
        return self.heat.resources.signal(stack_id, rsc_name)

    def resource_get(self, stack_id, rsc_name):
        return self.heat.resources.get(stack_id, rsc_name)

    def resource_event_list(self, stack_id, rsc_name, **kwargs):
        return self.heat.events.list(stack_id, rsc_name, **kwargs)

    def resource_metadata(self, stack_id, rsc_name):
        return self.heat.resources.metadata(stack_id, rsc_name)

    def resource_mark_unhealthy(self, stack_id, resource_name,
                                mark_unhealthy, resource_status_reason):
        return self.heat.resources.mark_unhealthy(stack_id, resource_name,
                                                  mark_unhealthy,
                                                  resource_status_reason)



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""change type for instance_id of vnf table

Revision ID: 7186440a306b
Revises: df26c5871f3c
Create Date: 2020-12-21 16:08:23.438871

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '7186440a306b'
down_revision = 'df26c5871f3c'


def upgrade(active_plugins=None, options=None):
    op.alter_column('vnf',
                    'instance_id',
                    type_=sa.String(length=12800),
                    nullable=True)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.sol_refactored.objects import base
from tacker.sol_refactored.objects import fields


# NFV-SOL 013
# - v3.4.1 Table 8.3.4-1
@base.TackerObjectRegistry.register
class SubscriptionAuthentication(base.TackerObject,
                                 base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'authType': fields.ListOfEnumField(
            valid_values=[
                'BASIC',
                'OAUTH2_CLIENT_CREDENTIALS',
                'TLS_CERT',
            ],
            nullable=False),
        'paramsBasic': fields.ObjectField(
            'SubscriptionAuthentication_ParamsBasic', nullable=True),
        'paramsOauth2ClientCredentials': fields.ObjectField(
            'SubscriptionAuthentication_ParamsOauth2', nullable=True),
    }


@base.TackerObjectRegistry.register
class SubscriptionAuthentication_ParamsBasic(base.TackerObject,
                                             base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'userName': fields.StringField(nullable=True),
        'password': fields.StringField(nullable=True),
    }


# NOTE: It should be
# SubscriptionAuthentication_ParamsOauth2ClientCredentials
# according to a principle, but shortened it as it is too long.

@base.TackerObjectRegistry.register
class SubscriptionAuthentication_ParamsOauth2(
        base.TackerObject, base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'clientId': fields.StringField(nullable=True),
        'clientPassword': fields.StringField(nullable=True),
        'tokenEndpoint': fields.UriField(nullable=True),
    }



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import io
import os
import shutil
import tempfile
import yaml
import zipfile

from oslo_log import log as logging

from tacker.sol_refactored.common import exceptions as sol_ex


LOG = logging.getLogger(__name__)


class Vnfd(object):

    def __init__(self, vnfd_id):
        self.vnfd_id = vnfd_id
        self.tosca_meta = {}
        self.definitions = {}
        self.vnfd_flavours = {}
        self.csar_dir = None
        self.csar_dir_is_tmp = False

    def init_from_csar_dir(self, csar_dir):
        self.csar_dir = csar_dir
        self.init_vnfd()

    def init_from_zip_file(self, zip_file):
        # NOTE: This is used when external NFVO is used.
        # TODO(oda-g): There is no delete route at the moment.
        # A possible enhance is that introducing cache management for
        # extracted vnf packages from external NFVO.
        self.csar_dir = tempfile.mkdtemp()
        self.csar_dir_is_tmp = True

        buff = io.BytesIO(zip_file)
        with zipfile.ZipFile(buff, 'r') as zf:
            zf.extractall(self.csar_dir)

        self.init_vnfd()

    def init_vnfd(self):
        # assume TOSCA-Metadata format
        path = os.path.join(self.csar_dir, 'TOSCA-Metadata', 'TOSCA.meta')
        if not os.path.isfile(path):
            raise sol_ex.InvalidVnfdFormat()

        # expand from yaml to dict for TOSCA.meta and Definitions
        with open(path, 'r') as f:
            self.tosca_meta = yaml.safe_load(f.read())

        path = os.path.join(self.csar_dir, 'Definitions')
        for entry in os.listdir(path):
            if entry.endswith(('.yaml', '.yml')):
                with open(os.path.join(path, entry), 'r') as f:
                    content = yaml.safe_load(f.read())
                self.definitions[entry] = content

    def delete(self):
        if self.csar_dir_is_tmp:
            shutil.rmtree(self.csar_dir)

    def get_vnfd_flavour(self, flavour_id):
        if flavour_id in self.vnfd_flavours:
            return self.vnfd_flavours[flavour_id]

        for data in self.definitions.values():
            fid = (data
                   .get('topology_template', {})
                   .get('substitution_mappings', {})
                   .get('properties', {})
                   .get('flavour_id'))
            if fid == flavour_id:
                self.vnfd_flavours[flavour_id] = data
                return data

        # NOT found.
        # NOTE: checked by the caller. basically check is necessary at
        # instantiate start only.

    def get_sw_image(self, flavour_id):
        vnfd = self.get_vnfd_flavour(flavour_id)
        nodes = (vnfd
                 .get('topology_template', {})
                 .get('node_templates', {}))
        types = ['tosca.nodes.nfv.Vdu.Compute',
                 'tosca.nodes.nfv.Vdu.VirtualBlockStorage']
        sw_image = {}
        for name, data in nodes.items():
            if (data['type'] in types and
                    data.get('properties', {}).get('sw_image_data')):
                image = data['properties']['sw_image_data']['name']
                sw_image[name] = image

        return sw_image

    def get_sw_image_data(self, flavour_id):
        vnfd = self.get_vnfd_flavour(flavour_id)
        nodes = (vnfd
                 .get('topology_template', {})
                 .get('node_templates', {}))
        types = ['tosca.nodes.nfv.Vdu.Compute',
                 'tosca.nodes.nfv.Vdu.VirtualBlockStorage']
        sw_image = {}
        for name, data in nodes.items():
            if (data['type'] in types and
                    data.get('properties', {}).get('sw_image_data')):
                sw_image[name] = data['properties']['sw_image_data']
                sw_file = (data
                           .get('artifacts', {})
                           .get('sw_image', {})
                           .get('file'))
                if sw_file:
                    sw_image[name]['file'] = sw_file

        return sw_image

    def get_vnfd_properties(self):
        """return properties used by instantiate"""
        # get from node_templates of VNF of
        # - ['properties']['configurable_properties']
        # - ['properties']['modifiable_attributes']['extensions']
        # - ['properties']['modifiable_attributes']['metadata']
        # NOTE: In etsi_nfv_sol001_vnfd_types.yaml which used by
        # tacker examples, definitions of these properties are commented out.

        prop = {
            'vnfConfigurableProperties': {},
            'extensions': {},
            'metadata': {}
        }
        return prop

    def get_nodes(self, flavour_id, node_type):
        vnfd = self.get_vnfd_flavour(flavour_id)
        nodes = (vnfd
                 .get('topology_template', {})
                 .get('node_templates', {}))

        res = {name: data
            for name, data in nodes.items() if data['type'] == node_type}

        return res

    def get_vdu_nodes(self, flavour_id):
        return self.get_nodes(flavour_id, 'tosca.nodes.nfv.Vdu.Compute')

    def get_storage_nodes(self, flavour_id):
        return self.get_nodes(flavour_id,
                              'tosca.nodes.nfv.Vdu.VirtualBlockStorage')

    def get_virtual_link_nodes(self, flavour_id):
        return self.get_nodes(flavour_id,
                              'tosca.nodes.nfv.VnfVirtualLink')

    def get_vducp_nodes(self, flavour_id):
        return self.get_nodes(flavour_id, 'tosca.nodes.nfv.VduCp')

    def get_vdu_cps(self, flavour_id, vdu_name):
        cp_nodes = self.get_vducp_nodes(flavour_id)
        cps = []
        for cp_name, cp_data in cp_nodes.items():
            reqs = cp_data.get('requirements', [])
            for req in reqs:
                if req.get('virtual_binding') == vdu_name:
                    cps.append(cp_name)
                    break
        return cps

    def get_vdu_storages(self, vdu_node):
        storages = [req['virtual_storage']
                    for req in vdu_node.get('requirements', [])
                    if 'virtual_storage' in req]

        return storages

    def get_base_hot(self, flavour_id):
        # NOTE: this method is openstack specific
        hot_dict = {}
        path = os.path.join(self.csar_dir, 'BaseHOT', flavour_id)
        if not os.path.isdir(path):
            return hot_dict

        for entry in os.listdir(path):
            if entry.endswith(('.yaml', '.yml')):
                with open(os.path.join(path, entry), 'r') as f:
                    content = yaml.safe_load(f.read())
                    hot_dict['template'] = content
                    break

        nested = os.path.join(path, 'nested')
        if not os.path.isdir(nested):
            return hot_dict

        for entry in os.listdir(nested):
            if entry.endswith(('.yaml', '.yml')):
                with open(os.path.join(nested, entry), 'r') as f:
                    content = yaml.safe_load(f.read())
                    hot_dict.setdefault('files', {})
                    hot_dict['files'][entry] = content

        return hot_dict

    def get_vl_name_from_cp(self, flavour_id, cp_data):
        for req in cp_data.get('requirements', []):
            if 'virtual_link' in req:
                return req['virtual_link']

    def get_compute_flavor(self, flavour_id, vdu_name):
        vnfd = self.get_vnfd_flavour(flavour_id)
        flavor = (vnfd.get('topology_template', {})
                      .get('node_templates', {})
                      .get(vdu_name, {})
                      .get('capabilities', {})
                      .get('virtual_compute', {})
                      .get('properties', {})
                      .get('requested_additional_capabilities', {})
                      .get('properties', {})
                      .get('requested_additional_capability_name'))
        if flavor:
            return flavor

    def make_tmp_csar_dir(self):
        # If this fails, 500 which is not caused by programming error
        # but true 'Internal server error' raises.
        tmp_dir = tempfile.mkdtemp()
        # remove tmp_dir because copytree fails if destination exists.
        # NOTE:
        # so mkdtemp is used for getting a unique name at the moment.
        # if py38 or later, copytree supports dirs_exits_ok parameter.
        # It is not necessary to remove tmp_dir, specify
        # 'dirs_exists_ok=True' to copytree instead when tacker support
        # only py38 or later.
        os.rmdir(tmp_dir)
        shutil.copytree(self.csar_dir, tmp_dir,
            ignore=shutil.ignore_patterns('Files'))
        return tmp_dir

    def remove_tmp_csar_dir(self, tmp_dir):
        try:
            shutil.rmtree(tmp_dir)
        except Exception:
            LOG.exception("rmtree %s failed", tmp_dir)
            # as this error does not disturb the process, continue.

    def get_policy_values_by_type(self, flavour_id, policy_type):
        vnfd = self.get_vnfd_flavour(flavour_id)
        policies = (vnfd.get('topology_template', {})
                        .get('policies', []))
        if isinstance(policies, dict):
            policies = [policies]

        ret = [value
            for policy in policies for value in policy.values()
            if value['type'] == policy_type]

        return ret

    def get_default_instantiation_level(self, flavour_id):
        policies = self.get_policy_values_by_type(flavour_id,
                'tosca.policies.nfv.InstantiationLevels')
        if policies:
            return policies[0].get('properties', {}).get('default_level')

    def get_vdu_num(self, flavour_id, vdu_name, instantiation_level):
        policies = self.get_policy_values_by_type(flavour_id,
                'tosca.policies.nfv.VduInstantiationLevels')
        for policy in policies:
            if vdu_name in policy.get('targets', []):
                return (policy.get('properties', {})
                              .get('levels', {})
                              .get(instantiation_level, {})
                              .get('number_of_instances'))
        return 0

    def get_placement_groups(self, flavour_id):
        vnfd = self.get_vnfd_flavour(flavour_id)
        groups = (vnfd.get('topology_template', {})
                      .get('groups', []))
        if isinstance(groups, dict):
            groups = [groups]

        ret = {key: value['members']
            for group in groups for key, value in group.items()
            if value['type'] == 'tosca.groups.nfv.PlacementGroup'}

        return ret

    def _get_targets(self, flavour_id, affinity_type):
        policies = self.get_policy_values_by_type(flavour_id, affinity_type)
        groups = self.get_placement_groups(flavour_id)

        ret = []
        for policy in policies:
            scope = policy['properties']['scope']
            if scope not in ['zone', 'nfvi_node']:
                continue

            targets = []
            for target in policy['targets']:
                if target in list(groups.keys()):
                    targets += groups[target]
                else:
                    targets.append(target)

            ret.append((targets, scope))

        return ret

    def get_affinity_targets(self, flavour_id):
        return self._get_targets(flavour_id,
                'tosca.policies.nfv.AffinityRule')

    def get_anti_affinity_targets(self, flavour_id):
        return self._get_targets(flavour_id,
                'tosca.policies.nfv.AntiAffinityRule')

    def get_interface_script(self, flavour_id, operation):
        vnfd = self.get_vnfd_flavour(flavour_id)
        nodes = (vnfd.get('topology_template', {})
                     .get('node_templates', {}))
        for node in nodes.values():
            if 'interfaces' not in node:
                continue
            op_value = (node['interfaces'].get('Vnflcm', {})
                                          .get(operation))
            if not isinstance(op_value, dict):
                # op_value may be []
                return

            artifact = op_value.get('implementation')
            if artifact is None:
                # no script specified for the operation
                return

            script = (node.get('artifacts', {})
                          .get(artifact, {})
                          .get('file'))
            if script is None:
                # can not happen if vnf package is correct.
                return

            script_type = node['artifacts'][artifact].get('type')
            if script_type != 'tosca.artifacts.Implementation.Python':
                # support python script only at the moment
                msg = "Unsupported script type {}".format(script_type)
                raise sol_ex.SolHttpError422(sol_detail=msg)

            return script

    def get_scale_vdu_and_num(self, flavour_id, aspect_id):
        aspects = self.get_policy_values_by_type(flavour_id,
            'tosca.policies.nfv.ScalingAspects')
        delta = None
        for aspect in aspects:
            value = aspect['properties']['aspects'].get(aspect_id)
            if value is not None:
                # expect there is one delta.
                # NOTE: Tacker does not support non-uniform deltas defined in
                # ETSI NFV SOL001 8. Therefore, uniform delta corresponding
                # to number_of_instances can be set and number_of_instances is
                # the same regardless of scale_level.
                delta = value['step_deltas'][0]
                break

        if delta is None:
            return {}

        aspect_deltas = self.get_policy_values_by_type(flavour_id,
            'tosca.policies.nfv.VduScalingAspectDeltas')
        vdu_num_inst = {}
        for aspect_delta in aspect_deltas:
            if aspect_delta.get('properties', {}).get('aspect') == aspect_id:
                num_inst = (aspect_delta['properties']['deltas']
                            .get(delta, {}).get('number_of_instances'))
                # NOTE: it is not checked whether 'delta' defined in
                # ScaleingAspects exists in VduScalingAspectDeltas at
                # the loading of vnf package. this is a mistake of the
                # VNFD definition.
                if num_inst is None:
                    raise sol_ex.DeltaMissingInVnfd(delta=delta)
                for vdu_name in aspect_delta['targets']:
                    vdu_num_inst[vdu_name] = num_inst

        return vdu_num_inst

    def get_scale_info_from_inst_level(self, flavour_id, inst_level):
        policies = self.get_policy_values_by_type(flavour_id,
            'tosca.policies.nfv.InstantiationLevels')
        for policy in policies:
            return (policy['properties']['levels']
                    .get(inst_level, {})
                    .get('scale_info', {}))
        return {}

    def get_max_scale_level(self, flavour_id, aspect_id):
        aspects = self.get_policy_values_by_type(flavour_id,
            'tosca.policies.nfv.ScalingAspects')
        for aspect in aspects:
            value = aspect['properties']['aspects'].get(aspect_id)
            if value is not None:
                return value['max_scale_level']

        # should not occur
        return 0



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""create of Network service tables

Revision ID: 0ad3bbce1c18
Revises: 0ae5b1ce3024
Create Date: 2016-12-17 19:41:01.906138

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '0ad3bbce1c18'
down_revision = '8f7145914cb0'

from alembic import op
import sqlalchemy as sa

from tacker.db import types


def upgrade(active_plugins=None, options=None):
    op.create_table('nsd',
        sa.Column('tenant_id', sa.String(length=64), nullable=False),
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('description', sa.Text(), nullable=True),
        sa.Column('vnfds', types.Json, nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_table('ns',
        sa.Column('tenant_id', sa.String(length=64), nullable=False),
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('nsd_id', types.Uuid(length=36), nullable=True),
        sa.Column('vim_id', sa.String(length=64), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('description', sa.Text(), nullable=True),
        sa.Column('vnf_ids', sa.TEXT(length=65535), nullable=True),
        sa.Column('mgmt_urls', sa.TEXT(length=65535), nullable=True),
        sa.Column('status', sa.String(length=64), nullable=False),
        sa.Column('error_reason', sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(['nsd_id'], ['nsd.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_table('nsd_attribute',
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('nsd_id', types.Uuid(length=36), nullable=False),
        sa.Column('key', sa.String(length=255), nullable=False),
        sa.Column('value', sa.TEXT(length=65535), nullable=True),
        sa.ForeignKeyConstraint(['nsd_id'], ['nsd.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )



# Copyright 2018 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import versionutils

from tacker.objects import base
from tacker.objects import fields


@base.TackerObjectRegistry.register
class HealVnfAdditionalParams(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'parameter': fields.StringField(),
        'cause': fields.ListOfStringsField()
    }


@base.TackerObjectRegistry.register
class HealVnfRequest(base.TackerObject):

    # Version 1.0: Initial version
    # Version 1.1: Added vnf_instance_id
    # Version 1.2: Added stack_id for nested heat-template
    VERSION = '1.2'

    fields = {
        'vnfc_instance_id': fields.ListOfStringsField(nullable=True,
                                                      default=[]),
        'stack_id': fields.StringField(nullable=True, default=''),
        'cause': fields.StringField(nullable=True, default=None),
        'legacy_additional_params': fields.ListOfObjectsField(
            'HealVnfAdditionalParams', default=[]),
        'additional_params': fields.DictOfNullableField(nullable=True,
                                                        default={})
    }

    def obj_make_compatible(self, primitive, target_version):
        super(HealVnfRequest, self).obj_make_compatible(primitive,
                                                        target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1) and 'vnfc_instance_id' in primitive:
            del primitive['vnfc_instance_id']



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import iso8601
import os
import shutil
from tacker import objects
from tacker.objects import fields
from tacker.tests import constants
import tempfile
import uuid
import yaml
import zipfile

from oslo_config import cfg

from tacker.common import utils as common_utils
from tacker.db.db_sqlalchemy import models
from tacker.objects import scale_vnf_request
from tacker.tests import utils
from tacker.tests import uuidsentinel

VNF_UPLOAD_VNF_PACKAGE_CONTENT = {
    'algorithm': 'sha512', 'created_at': '2019-08-16T06:57:09Z',
    'deleted': False, 'deleted_at': None,
    'hash': 'ce48b8ba15bfb060fb70471cf955bef433e4513973b4bac42b37c36f57357dc35'
            'bf788c16545d3a59781914adf19fca26d6984583b7739e55c447383d774356a',
    'id': uuidsentinel.tenant_id,
    'location_glance_store': 'file:///var/lib/glance/images/'
                             'd617ea52-b16b-417e-a68c-08dfb69aab9e',
    'onboarding_state': 'PROCESSING', 'operational_state': 'DISABLED',
    'tenant_id': uuidsentinel.tenant_id,
    'updated_at': '2019-08-16T06:57:30Z',
    'usage_state': 'NOT_IN_USE', 'user_data': {'abc': 'xyz'}}

VNF_DATA = {
    'created_at': '2019-08-16T06:57:09Z',
    'deleted': False, 'deleted_at': None,
    'id': uuidsentinel.id,
    'onboarding_state': 'UPLOADING',
    'operational_state': 'DISABLED',
    'tenant_id': uuidsentinel.tenant_id,
    'updated_at': '2019-08-16T06:57:30Z',
    'usage_state': 'NOT_IN_USE',
    'user_data': {'abc': 'xyz'}
}

VNF_PACKAGE_DATA = {'algorithm': None, 'hash': None,
                    'location_glance_store': None,
                    'onboarding_state': 'CREATED',
                    'operational_state': 'DISABLED',
                    'tenant_id': uuidsentinel.tenant_id,
                    'usage_state': 'NOT_IN_USE',
                    'user_data': {'abc': 'xyz'}
                    }


def make_vnfd_files_list(csar_path):
    files_list = []
    # Checking for directory exist
    if not os.path.isdir(csar_path):
        return
    ext = ['.yaml', '.meta']
    for _, _, files in os.walk(csar_path):
        for file in files:
            if file.endswith(tuple(ext)):
                files_list.append(file)

    return files_list


def create_fake_csar_dir(vnf_package_id, temp_dir,
                         csar_without_tosca_meta=False):
    csar_dir = ('sample_vnfpkg_no_meta_single_vnfd' if csar_without_tosca_meta
                else 'sample_vnfpkg_tosca_vnfd')
    fake_csar = os.path.join(temp_dir, vnf_package_id)
    cfg.CONF.set_override('vnf_package_csar_path', temp_dir,
                          group='vnf_package')
    utils.copy_csar_files(fake_csar, csar_dir, csar_without_tosca_meta)

    return fake_csar


def get_vnf_package_vnfd():
    return {
        "id": uuidsentinel.vnfd_id,
        "vnf_provider": "test vnf provider",
        "vnf_product_name": "Sample VNF",
        "vnf_software_version": "1.0",
        "vnfd_version": "1.0",
        "name": 'Sample VNF Instance',
    }


def get_lcm_op_occs_data():
    return {
        "id": uuidsentinel.lcm_op_occs_id,
        "tenant_id": uuidsentinel.tenant_id,
        'operation_state': 'PROCESSING',
        'state_entered_time':
        datetime.datetime(1900, 1, 1, 1, 1, 1,
                          tzinfo=iso8601.UTC),
        'start_time': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                        tzinfo=iso8601.UTC),
        'operation': 'MODIFY_INFO',
        'is_automatic_invocation': 0,
        'is_cancel_pending': 0,
    }


def get_vnf_lcm_subscriptions():
    subscription_id = uuidsentinel.subscription_id
    return {
        "id": subscription_id.encode(),
        "callback_uri": b'http://localhost:9890/'
    }


def get_expected_vnfd_data(zip_file=None):
    if zip_file:
        csar_temp_dir = tempfile.mkdtemp()
        with zipfile.ZipFile(zip_file, 'r') as zf:
            zf.extractall(csar_temp_dir)
    else:
        unique_name = str(uuid.uuid4())
        csar_temp_dir = os.path.join('/tmp', unique_name)
        utils.copy_csar_files(csar_temp_dir, 'sample_vnfpkg_tosca_vnfd',
                              read_vnfd_only=True)

    file_names = ['TOSCA-Metadata/TOSCA.meta',
                  'Definitions/etsi_nfv_sol001_vnfd_types.yaml',
                  'Definitions/helloworld3_types.yaml',
                  'Definitions/helloworld3_df_simple.yaml',
                  'Definitions/helloworld3_top.vnfd.yaml',
                  'Definitions/etsi_nfv_sol001_common_types.yaml']
    file_path_and_data = {}
    for file_name in file_names:
        with open(os.path.join(csar_temp_dir, file_name)) as f:
            file_path_and_data.update({file_name: yaml.dump(
                yaml.safe_load(f))})

    shutil.rmtree(csar_temp_dir)
    return file_path_and_data


def fake_vnf_package_vnfd_model_dict(**updates):
    vnfd = {
        'package_uuid': uuidsentinel.package_uuid,
        'deleted': False,
        'deleted_at': None,
        'updated_at': None,
        'created_at': datetime.datetime(2020, 1, 1, 1, 1, 1,
                                        tzinfo=iso8601.UTC),
        'vnf_product_name': 'Sample VNF',
        'vnf_provider': 'test vnf provider',
        'vnf_software_version': '1.0',
        'vnfd_id': uuidsentinel.vnfd_id,
        'vnfd_version': '1.0',
        'id': constants.UUID,
    }

    if updates:
        vnfd.update(updates)

    return vnfd


def return_vnf_package_vnfd():
    model_obj = models.VnfPackageVnfd()
    model_obj.update(fake_vnf_package_vnfd_model_dict())
    return model_obj


def _model_non_instantiated_vnf_instance(**updates):
    vnf_instance = {
        'created_at': datetime.datetime(2020, 1, 1, 1, 1, 1,
                                        tzinfo=iso8601.UTC),
        'deleted': False,
        'deleted_at': None,
        'id': uuidsentinel.vnf_instance_id,
        'instantiated_vnf_info': None,
        'instantiation_state': fields.VnfInstanceState.NOT_INSTANTIATED,
        'updated_at': None,
        'vim_connection_info': [],
        'vnf_instance_description': 'Vnf instance description',
        'vnf_instance_name': 'Vnf instance name',
        'vnf_product_name': 'Sample VNF',
        'vnf_provider': 'Vnf provider',
        'vnf_software_version': '1.0',
        'tenant_id': uuidsentinel.tenant_id,
        'vnfd_id': uuidsentinel.vnfd_id,
        'vnfd_version': '1.0',
        'vnfPkgId': uuidsentinel.vnf_pkg_id}

    if updates:
        vnf_instance.update(**updates)

    return vnf_instance


def return_vnf_instance(
        instantiated_state=fields.VnfInstanceState.NOT_INSTANTIATED,
        scale_status=None,
        **updates):

    if instantiated_state == fields.VnfInstanceState.NOT_INSTANTIATED:
        data = _model_non_instantiated_vnf_instance(**updates)
        data['instantiation_state'] = instantiated_state
        vnf_instance_obj = objects.VnfInstance(**data)

    elif scale_status:
        data = _model_non_instantiated_vnf_instance(**updates)
        data['instantiation_state'] = instantiated_state
        vnf_instance_obj = objects.VnfInstance(**data)

        get_instantiated_vnf_info = {
            'flavour_id': uuidsentinel.flavour_id,
            'vnf_state': 'STARTED',
            'instance_id': uuidsentinel.instance_id
        }
        instantiated_vnf_info = get_instantiated_vnf_info

        s_status = {"aspect_id": "SP1", "scale_level": 1}
        scale_status = objects.ScaleInfo(**s_status)

        instantiated_vnf_info.update(
            {"ext_cp_info": [],
             'ext_virtual_link_info': [],
             'ext_managed_virtual_link_info': [],
             'vnfc_resource_info': [],
             'vnf_virtual_link_resource_info': [],
             'virtual_storage_resource_info': [],
             "flavour_id": "simple",
             "scale_status": [scale_status],
             "vnf_instance_id": "171f3af2-a753-468a-b5a7-e3e048160a79",
             "additional_params": {"key": "value"},
             'vnf_state': "STARTED"})
        info_data = objects.InstantiatedVnfInfo(**instantiated_vnf_info)

        vnf_instance_obj.instantiated_vnf_info = info_data
    else:
        data = _model_non_instantiated_vnf_instance(**updates)
        data['instantiation_state'] = instantiated_state
        vnf_instance_obj = objects.VnfInstance(**data)
        inst_vnf_info = objects.InstantiatedVnfInfo.obj_from_primitive({
            "ext_cp_info": [],
            'ext_virtual_link_info': [],
            'ext_managed_virtual_link_info': [],
            'vnfc_resource_info': [],
            'vnf_virtual_link_resource_info': [],
            'virtual_storage_resource_info': [],
            "flavour_id": "simple",
            "additional_params": {"key": "value"},
            'vnf_state': "STARTED"}, None)

        vnf_instance_obj.instantiated_vnf_info = inst_vnf_info

    return vnf_instance_obj


def _get_vnf(**updates):
    vnf_data = {
        'tenant_id': uuidsentinel.tenant_id,
        'name': "fake_name",
        'vnfd_id': uuidsentinel.vnfd_id,
        'vnf_instance_id': uuidsentinel.instance_id,
        'mgmt_ip_address': "fake_mgmt_ip_address",
        'status': 'ACTIVE',
        'description': 'fake_description',
        'placement_attr': 'fake_placement_attr',
        'vim_id': 'uuidsentinel.vim_id',
        'error_reason': 'fake_error_reason',
    }

    if updates:
        vnf_data.update(**updates)

    return vnf_data


def scale_request(type, number_of_steps):
    scale_request_data = {
        'type': type,
        'aspect_id': "SP1",
        'number_of_steps': number_of_steps,
        'scale_level': 1,
        'additional_params': {"test": "test_value"},
    }
    scale_request = scale_vnf_request.ScaleVnfRequest(**scale_request_data)

    return scale_request


def get_instantiated_vnf_info():
    vnf_info = objects.vnf_instance.VnfInstance()

    def _get_instantiated_vnf_info_data():
        return {
            "flavour_id": "simple",
            "vnf_instance_id": uuidsentinel.vnf_instance_id,
            "vnf_virtual_link_resource_info":
                _get_virtual_link_resource_info(),
            "vnfc_resource_info": _get_vnfc_resource_info(),
        }

    def _get_virtual_link_resource_info():
        return [{
            "id": uuidsentinel.vnf_vl_resource_1,
            "vnf_virtual_link_desc_id": uuidsentinel.ext_vl_1,
            "network_resource": {
                "vim_connection_id": None,
                "resource_id": uuidsentinel.ext_vl_resource_1,
                "vim_level_resource_type": "OS::Neutron::Net",
                "deleted": False,
            },
            "vnf_link_ports": [
                {
                    "id": uuidsentinel.vnf_link_port_1,
                    "resource_handle": {
                        "vim_connection_id": uuidsentinel.vim_connection_id,
                        "resource_id": uuidsentinel.vnf_vl1_link_port_1,
                        "vim_level_resource_type": "OS::Neutron::Port",
                        "deleted": False,
                    },
                    "cp_instance_id": uuidsentinel.cp_instance_1,
                    "deleted": False,
                },
                {
                    "id": uuidsentinel.vnf_link_port_2,
                    "resource_handle": {
                        "vim_connection_id": uuidsentinel.vim_connection_id,
                        "resource_id": uuidsentinel.vnf_vl1_link_port_2,
                        "vim_level_resource_type": "OS::Neutron::Port",
                        "deleted": False,
                    },
                    "cp_instance_id": uuidsentinel.cp_instance_2,
                    "deleted": False,
                },
                {
                    "id": uuidsentinel.vnf_link_port_5,
                    "resource_handle": {
                        "vim_connection_id": uuidsentinel.vim_connection_id,
                        "resource_id": uuidsentinel.vnf_vl1_link_port_5,
                        "vim_level_resource_type": "OS::Neutron::Port",
                        "deleted": False,
                    },
                    "cp_instance_id": uuidsentinel.cp_instance_2,
                    "deleted": False,
                },
            ],
        },
            {
            "id": uuidsentinel.vnf_vl_resource_2,
            "vnf_virtual_link_desc_id": uuidsentinel.ext_vl_2,
            "network_resource": {
                "vim_connection_id": None,
                "resource_id": uuidsentinel.ext_vl_resource_2,
                "vim_level_resource_type": "OS::Neutron::Net",
                "deleted": False,
            },
            "vnf_link_ports": [
                {
                    "id": uuidsentinel.vnf_link_port_3,
                    "resource_handle": {
                        "vim_connection_id": uuidsentinel.vim_connection_id,
                        "resource_id": uuidsentinel.vnf_vl2_link_port_1,
                        "vim_level_resource_type": "OS::Neutron::Port",
                        "deleted": False,
                    },
                    "cp_instance_id": uuidsentinel.cp_instance_3,
                    "deleted": False,
                },
                {
                    "id": uuidsentinel.vnf_link_port_4,
                    "resource_handle": {
                        "vim_connection_id": uuidsentinel.vim_connection_id,
                        "resource_id": uuidsentinel.vnf_vl2_link_port_2,
                        "vim_level_resource_type": "OS::Neutron::Port",
                        "deleted": False,
                    },
                    "cp_instance_id": uuidsentinel.cp_instance_4,
                    "deleted": False,
                },
                {
                    "id": uuidsentinel.vnf_link_port_5,
                    "resource_handle": {
                        "vim_connection_id": uuidsentinel.vim_connection_id,
                        "resource_id": uuidsentinel.vnf_vl2_link_port_3,
                        "vim_level_resource_type": "OS::Neutron::Port",
                        "deleted": False,
                    },
                    "cp_instance_id": uuidsentinel.cp_instance_5,
                    "deleted": False,
                }
            ]
        }]

    def _get_vnfc_resource_info():
        return [{
            "id": uuidsentinel.vnfc_resource_1,
            "vdu_id": "VDU1",
            "compute_resource": {
                "vim_connection_id": uuidsentinel.vim_connection_id,
                "resource_id": uuidsentinel.uuid,
                "vim_level_resource_type": "OS::Nova::Server",
                "deleted": False,
            },
            "storage_resource_ids": [],
            "vnfc_cp_info": [{
                "id": uuidsentinel.cp_instance_1,
                "cpd_id": "CP1",
                "vnf_ext_cp_id": None,
                "cp_protocol_info": [],
                "vnf_link_port_id": uuidsentinel.vnf_link_port_1,
            }, {
                "id": uuidsentinel.cp_instance_2,
                "cpd_id": "CP2",
                "vnf_ext_cp_id": None,
                "cp_protocol_info": [],
                "vnf_link_port_id": uuidsentinel.vnf_link_port_2,
            }],
            "metadata": {},
        }, {
            "id": uuidsentinel.vnfc_resource_2,
            "vdu_id": "VDU2",
            "compute_resource": {
                "vim_connection_id": uuidsentinel.vim_connection_id,
                "resource_id": uuidsentinel.uuid,
                "vim_level_resource_type": "OS::Nova::Server",
                "deleted": False,
            },
            "storage_resource_ids": [],
            "vnfc_cp_info": [{
                "id": uuidsentinel.cp_instance_3,
                "cpd_id": "CP3",
                "vnf_ext_cp_id": None,
                "cp_protocol_info": [],
                "vnf_link_port_id": uuidsentinel.vnf_link_port_3,
            }, {
                "id": uuidsentinel.cp_instance_4,
                "cpd_id": "CP4",
                "vnf_ext_cp_id": None,
                "cp_protocol_info": [],
                "vnf_link_port_id": uuidsentinel.vnf_link_port_4,
            }, {
                "id": uuidsentinel.cp_instance_5,
                "cpd_id": "CP5",
                "vnf_ext_cp_id": None,
                "cp_protocol_info": [],
                "vnf_link_port_id": uuidsentinel.vnf_link_port_5,
            }],
            "metadata": {},
        }]

    vnf_info.instantiated_vnf_info = \
        objects.InstantiatedVnfInfo.obj_from_primitive(
            _get_instantiated_vnf_info_data(), None
        )
    return vnf_info.instantiated_vnf_info


def get_change_ext_conn_request():
    change_ext_conn_req_body = {
        "extVirtualLinks": [{
            "id": uuidsentinel.ext_vl_1,
            "vimConnectionId": uuidsentinel.vim_connection_id,
            "resourceId": uuidsentinel.ext_vl_resource_3,
            "extCps": [{
                "cpdId": 'CP1',
                "cpConfig": [{
                    "cpInstanceId": uuidsentinel.uuid,
                    "cpProtocolData": [{
                        "layerProtocol": 'IP_OVER_ETHERNET',
                        "ipOverEthernet": {
                            "ipAddresses": [{
                                "type": "IPV4",
                                "fixedAddresses": ["10.0.0.1"],
                                "subnetId": uuidsentinel.uuid,
                            }]
                        }
                    }]
                }]
            }]}, {
            "id": uuidsentinel.ext_vl_2,
            "vimConnectionId": uuidsentinel.vim_connection_id,
            "resourceId": uuidsentinel.ext_vl_resource_2,
            "extCps": [{
                "cpdId": 'CP3',
                "cpConfig": [{
                    "cpInstanceId": uuidsentinel.uuid,
                    "cpProtocolData": [{
                        "layerProtocol": 'IP_OVER_ETHERNET',
                        "ipOverEthernet": {
                            "ipAddresses": [{
                                "type": "IPV4",
                                "fixedAddresses": ["10.0.0.2"],
                                "subnetId": uuidsentinel.uuid,
                            }]
                        }
                    }]
                }]}, {
                "cpdId": 'CP5',
                "cpConfig": [{
                    "cpInstanceId": uuidsentinel.uuid,
                    "cpProtocolData": [{
                        "layerProtocol": 'IP_OVER_ETHERNET',
                        "ipOverEthernet": {
                            "ipAddresses": [{
                                "type": "IPV4",
                                "numDynamicAddresses": 1,
                                "subnetId": uuidsentinel.uuid,
                            }]
                        }
                    }]
                }]
            }]
        }],
        "vimConnectionInfo": [{
            "id": uuidsentinel.vim_connection_id,
            "vimId": uuidsentinel.uuid,
            "vimType": 'openstack',
            "interfaceInfo": {"key1": 'value1', "key2": 'value2'},
            "accessInfo": {"key1": 'value1', "key2": 'value2'},
        }],
    }

    return change_ext_conn_req_body


def get_change_ext_conn_request_obj():
    """Return ChangeExtConnRequest Object

    obj_from_primitive() needs snake_case dictionary
    """
    body = common_utils.convert_camelcase_to_snakecase(
        get_change_ext_conn_request())
    return objects.ChangeExtConnRequest.obj_from_primitive(
        body, None)



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""unique_constraint_name

Revision ID: 0ae5b1ce3024
Revises: 507122918800
Create Date: 2016-09-15 16:27:08.736673

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '0ae5b1ce3024'
down_revision = '507122918800'

from alembic import op
import sqlalchemy as sa


def _migrate_duplicate_names(table):

    meta = sa.MetaData(bind=op.get_bind())
    t = sa.Table(table, meta, autoload=True)

    session = sa.orm.Session(bind=op.get_bind())
    with session.begin(subtransactions=True):
        dup_names = session.query(t.c.name).group_by(
            t.c.name).having(sa.func.count() > 1).all()
        if dup_names:
            for name in dup_names:
                duplicate_obj_query = session.query(t).filter(t.c.name == name[
                    0]).all()
                for dup_obj in duplicate_obj_query:
                    name = dup_obj.name[:242] if dup_obj.name > 242 else \
                        dup_obj.name
                    new_name = '{0}-{1}'.format(name, dup_obj.id[-12:])
                    session.execute(t.update().where(
                        t.c.id == dup_obj.id).values(name=new_name))
    session.commit()


def upgrade(active_plugins=None, options=None):

    _migrate_duplicate_names('vnf')
    _migrate_duplicate_names('vnfd')
    _migrate_duplicate_names('vims')
